/*
    MiddleMan filtering proxy server
    Copyright (C) 2002  Jason McLaughlin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include "proto.h"

extern GLOBAL *global;

#ifdef HAVE_ZLIB
#include <zlib.h>
#endif				/* HAVE_ZLIB */

/*
free memory used by FILEBUF-type structure
*/
void filebuf_free(FILEBUF * filebuf)
{
	if (!filebuf)
		return;

	if (filebuf->data != NULL)
		xfree(filebuf->data);
	if (filebuf->type != NULL)
		xfree(filebuf->type);

	xfree(filebuf);
}

/*
create empty filebuf
*/
FILEBUF *filebuf_new()
{
	FILEBUF *filebuf;

	filebuf = xmalloc(sizeof(FILEBUF));

	filebuf->data = NULL;
	filebuf->type = NULL;
	filebuf->size = filebuf->realsize = 0;
	filebuf->code = -1;

	return filebuf;
}

FILEBUF *filebuf_dup(FILEBUF * filebuf)
{
	FILEBUF *ret;

	ret = filebuf_new();

	ret->data = (filebuf->size != 0) ? xmalloc(filebuf->size) : NULL;
	memcpy(ret->data, filebuf->data, filebuf->size);
	ret->size = ret->realsize = filebuf->size;

	return ret;
}

/*
reduce filebuf size to actual size of contents
*/
void filebuf_shorten(FILEBUF * filebuf)
{
	if (filebuf->realsize > filebuf->size) {
		filebuf->realsize = filebuf->size;
		filebuf->data = xrealloc(filebuf->data, filebuf->size);
	}
}

/*
append data to a file buffer
*/
void filebuf_add(FILEBUF * filebuf, const char *data, int len)
{
	int x;

	x = filebuf->size;

	filebuf_resize(filebuf, filebuf->size + len);

	memcpy(&filebuf->data[x], data, len);
}


/*
add formatted data to filebuf
*/
void filebuf_addf(FILEBUF * filebuf, const char *fmt, ...)
{
	int ret;
	char buf[8192];
	va_list valist;

	va_start(valist, fmt);
	ret = vsnprintf(buf, sizeof(buf), fmt, valist);
	va_end(valist);

	filebuf_add(filebuf, buf, (ret > sizeof(buf) || ret == -1) ? sizeof(buf) : ret);
}

/*
resize a filebuf (without adding anything)
*/
void filebuf_resize(FILEBUF * filebuf, int newsize)
{
	if (filebuf->realsize >= newsize && filebuf->realsize <= newsize + FILEBUF_ALIGNMENT && filebuf->data != NULL)
		filebuf->size = newsize;
	else if (newsize != 0) {
		/* aligning the filebuf size makes it much more difficult to catch off-by-one errors */
#ifdef _DEBUG
		filebuf->realsize = newsize;
#else
		filebuf->realsize = ALIGN2(newsize, FILEBUF_ALIGNMENT);
#endif
		filebuf->data = xrealloc(filebuf->data, filebuf->realsize);
		filebuf->size = newsize;
	}
}

/*
write a filebuf to a file
*/
int filebuf_save(FILEBUF * filebuf, char *filename)
{
	int fd, ret;

	fd = open(filename, O_CREAT | O_WRONLY);
	if (fd == -1)
		return -1;

	ret = write(fd, filebuf->data, filebuf->size);

	close(fd);

	return ret;
}

/*
read a file into a filebuf
*/
int filebuf_read(FILEBUF * filebuf, char *filename)
{
	int x, ret = 0, fd;
	char buf[BLOCKSIZE];

	fd = open(filename, O_RDONLY);
	if (fd == -1)
		return -1;

	while ((x = read(fd, buf, BLOCKSIZE)) > 0) {
		filebuf_add(filebuf, buf, x);

		ret += x;
	}

	close(fd);

	return ret;
}

/*
replace area of filebuf at specified offset and length with something else
*/
void filebuf_replace(FILEBUF * filebuf, int so, int eo, char *replace)
{
	int len1 = eo - so;
	int len2 = strlen(replace);

	if (len2 > len1)
		filebuf_resize(filebuf, filebuf->size + (len2 - len1));

	if (len1 != len2)
		memmove(&filebuf->data[eo + (len2 - len1)], &filebuf->data[eo], filebuf->size - eo);

	if (len2 < len1)
		filebuf_resize(filebuf, filebuf->size + (len2 - len1));

	memcpy(&filebuf->data[so], replace, len2);
}

#ifdef HAVE_ZLIB

int filebuf_compress(FILEBUF * filebuf, int gzip)
{
	int x, i, done = 0;
	char gzhdr[10] = { 0x1f, 0x8b, Z_DEFLATED, 0, 0, 0, 0, 0, 0, 0x03 };
	char gzbuffer[8192], *ptr;
	unsigned long gz_crc32, len;
	z_stream gz_stream;
	FILEBUF *cfilebuf = NULL;

	/* it seems zlib makes an allocation with a size of 0, so using the xmalloc/xfree
	   wrappers is out of the question since they assert for that */
	gz_stream.zalloc = NULL;
	gz_stream.zfree = NULL;
	gz_stream.opaque = NULL;
	gz_stream.next_in = gz_stream.next_out = 0;
	gz_stream.avail_in = gz_stream.avail_out = 0;

	x = deflateInit2(&gz_stream, Z_BEST_COMPRESSION, Z_DEFLATED, -MAX_WBITS, 9, Z_DEFAULT_STRATEGY);
	if (x != Z_OK)
		return FALSE;

	cfilebuf = filebuf_new();
	if (gzip == TRUE)
		filebuf_add(cfilebuf, gzhdr, sizeof(gzhdr));

	gz_stream.next_out = gzbuffer;
	gz_stream.avail_out = sizeof(gzbuffer);
	gz_stream.next_in = filebuf->data;
	gz_stream.avail_in = filebuf->size;

	while (gz_stream.avail_in != 0) {
		if (gz_stream.avail_out == 0) {
			filebuf_add(cfilebuf, gzbuffer, sizeof(gzbuffer));

			gz_stream.next_out = gzbuffer;
			gz_stream.avail_out = sizeof(gzbuffer);
		}

		x = deflate(&gz_stream, Z_NO_FLUSH);
		if (x == Z_STREAM_END)
			break;
		else if (x != Z_OK)
			goto error;
	}

	while (1) {
		len = sizeof(gzbuffer) - gz_stream.avail_out;
		if (len > 0) {
			filebuf_add(cfilebuf, gzbuffer, len);

			gz_stream.next_out = gzbuffer;
			gz_stream.avail_out = sizeof(gzbuffer);
		}

		if (done)
			break;

		x = deflate(&gz_stream, Z_FINISH);
		if (len == 0 && x == Z_BUF_ERROR)
			x = Z_OK;

		done = (gz_stream.avail_out != 0 || x == Z_STREAM_END);

		if (x != Z_OK && x != Z_STREAM_END)
			break;
	}

	if (x != Z_STREAM_END)
		goto error;

	if (gzip == TRUE) {
		len = cfilebuf->size;
		filebuf_resize(cfilebuf, cfilebuf->size + 8);

		/* append the crc32 and size */
		gz_crc32 = crc32(0L, filebuf->data, filebuf->size);

		for (ptr = cfilebuf->data + len, i = 0; i < 4; i++) {
			*ptr++ = (gz_crc32 & 0xFF);
			gz_crc32 >>= 8;
		}

		x = gz_stream.total_in;
		for (i = 0; i < 4; i++) {
			*ptr++ = (x & 0xFF);
			x >>= 8;
		}
	}

	xfree(filebuf->data);
	filebuf->data = cfilebuf->data;
	filebuf->size = cfilebuf->size;
	filebuf->realsize = cfilebuf->realsize;

	cfilebuf->data = NULL;
	filebuf_free(cfilebuf);

	deflateEnd(&gz_stream);

	return TRUE;

      error:

	zlib_error(x);

	filebuf_free(cfilebuf);
	deflateEnd(&gz_stream);

	return FALSE;
}

int filebuf_uncompress(FILEBUF * filebuf, int gzip)
{
	int x, done = 0;
	char gzbuffer[8192];
	unsigned long len;
	z_stream gz_stream;
	FILEBUF *cfilebuf = NULL;

	/* sanity check, we assume it's at least this big further down */
	if (filebuf->size < 18 && gzip == TRUE)
		return FALSE;

	gz_stream.zalloc = NULL;
	gz_stream.zfree = NULL;
	gz_stream.opaque = NULL;
	gz_stream.next_in = gz_stream.next_out = 0;
	gz_stream.avail_in = gz_stream.avail_out = 0;

	x = inflateInit2(&gz_stream, -MAX_WBITS);
	if (x != Z_OK)
		return FALSE;

	cfilebuf = filebuf_new();

	gz_stream.next_out = gzbuffer;
	gz_stream.avail_out = sizeof(gzbuffer);
	if (gzip == TRUE) {
		/* skip gzip header */
		gz_stream.next_in = filebuf->data + 10;
		gz_stream.avail_in = filebuf->size - 18;
	} else {
		gz_stream.next_in = filebuf->data;
		gz_stream.avail_in = filebuf->size;
	}

	while (gz_stream.avail_in != 0) {
		if (gz_stream.avail_out == 0) {
			filebuf_add(cfilebuf, gzbuffer, sizeof(gzbuffer));

			gz_stream.next_out = gzbuffer;
			gz_stream.avail_out = sizeof(gzbuffer);
		}

		x = inflate(&gz_stream, Z_NO_FLUSH);
		if (x == Z_STREAM_END) 
			break;
		else if (x != Z_OK)
			goto error;
	}

	while (1) {
		len = sizeof(gzbuffer) - gz_stream.avail_out;
		if (len > 0) {
			filebuf_add(cfilebuf, gzbuffer, len);

			gz_stream.next_out = gzbuffer;
			gz_stream.avail_out = sizeof(gzbuffer);
		}

		if (done)
			break;

		x = inflate(&gz_stream, Z_SYNC_FLUSH);
		if (len == 0 && x == Z_BUF_ERROR)
			x = Z_OK;

		done = (gz_stream.avail_out != 0 || x == Z_STREAM_END);

		if (x != Z_OK && x != Z_STREAM_END)
			break;
	}

	if (x != Z_STREAM_END)
		goto error;

	/* skip validating crc32 and size... I'm lazy :P */

	xfree(filebuf->data);
	filebuf->data = cfilebuf->data;
	filebuf->size = cfilebuf->size;
	filebuf->realsize = cfilebuf->realsize;

	cfilebuf->data = NULL;
	filebuf_free(cfilebuf);

	inflateEnd(&gz_stream);

	return TRUE;

      error:

	zlib_error(x);

	filebuf_free(cfilebuf);
	inflateEnd(&gz_stream);

	return FALSE;
}

void zlib_error(int x)
{
	switch (x) {
	case Z_DATA_ERROR:
		putlog(MMLOG_ERROR, "zlib data error");
		break;
	case Z_MEM_ERROR:
		putlog(MMLOG_ERROR, "zlib memory error");
		break;
	case Z_BUF_ERROR:
		putlog(MMLOG_ERROR, "zlib buffer error");
		break;
	case Z_STREAM_ERROR:
		putlog(MMLOG_ERROR, "zlib stream error");
		break;
	case Z_NEED_DICT:
		putlog(MMLOG_ERROR, "zlib dictionary error");
		break;
	default:
		putlog(MMLOG_ERROR, "zlib unknown error %d", x);
		break;
	}
}

#else

int filebuf_compress(FILEBUF * filebuf, int gzip)
{
	return FALSE;
}

int filebuf_uncompress(FILEBUF * filebuf, int gzip)
{
	return FALSE;
}

#endif				/* HAVE_ZLIB */

STACK *filebuf_to_stack(FILEBUF * filebuf)
{
	char *lptr, *eolptr;
	STACK *ret = NULL, *start = NULL;

	lptr = filebuf->data;
	while (lptr != NULL) {
		eolptr = memchr(lptr, '\n', filebuf->size - (lptr - filebuf->data));

		PUSHN(ret, lptr, (eolptr != NULL) ? eolptr - lptr : filebuf->size - (lptr - filebuf->data));
		if (start == NULL)
			start = ret;

		lptr = (eolptr != NULL && eolptr - filebuf->data != filebuf->size - 1) ? eolptr + 1 : NULL;
	}

	return start;
}

int filebuf_decode(CONNECTION * connection, FILEBUF * filebuf)
{
	int ret;

	if (s_strcasestr(connection->rheader->content_encoding, "gzip") != NULL) {
		ret = filebuf_uncompress(filebuf, TRUE);
		if (ret == FALSE) {
			putlog(MMLOG_ERROR, "failed to decompress gzip encoded content");

			/* failed to decompress, just send as-is */
			return FALSE;
		} else
			putlog(MMLOG_DEBUG, "decompressed gzip encoded file");
	} else if (s_strcasestr(connection->rheader->content_encoding, "deflate") != NULL) {
		ret = filebuf_uncompress(filebuf, FALSE);
		if (ret == FALSE) {
			putlog(MMLOG_ERROR, "failed to decompress deflate encoded content");

			return FALSE;
		} else
			putlog(MMLOG_DEBUG, "decompressed deflate encoded file");
	} else {
		/* unhandled content encoding, send as-is */
		putlog(MMLOG_WARN, "unhandled content encoding: %s", connection->rheader->content_encoding);

		return FALSE;
	}

	return TRUE;
}

int filebuf_encode(CONNECTION * connection, FILEBUF * filebuf)
{
	int ret;

	if (connection->flags & CONNECTION_ENCODED)
		return FALSE;

	if (connection->header->accept_encoding != NULL && atomic_read_rwlock(&global->general->lock, &global->general->compressout) == TRUE) {
		/* compress the content if the browser supports gzip encoding */
		if (s_strcasestr(connection->header->accept_encoding, "gzip") != NULL) {
			ret = filebuf_compress(filebuf, TRUE);
			if (ret == TRUE) {
				putlog(MMLOG_DEBUG, "gzip encoded file");
				connection->rheader->content_encoding = xstrdup("gzip");

				return TRUE;
			} else {
				putlog(MMLOG_DEBUG, "failed to gzip encode file");

				return FALSE;
			}
		} else if (s_strcasestr(connection->header->accept_encoding, "deflate") != NULL) {
			ret = filebuf_compress(filebuf, FALSE);
			if (ret == TRUE) {
				putlog(MMLOG_DEBUG, "deflate encoded file");
				connection->rheader->content_encoding = xstrdup("deflate");

				return TRUE;
			} else {
				putlog(MMLOG_DEBUG, "failed to deflate encode file");

				return FALSE;
			}
		}
	}

	return FALSE;
}
