/*
    MiddleMan filtering proxy server
    Copyright (C) 2002  Jason McLaughlin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <errno.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/tcp.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "proto.h"

extern GLOBAL *global;

struct proxy_candidate *plist_add(struct proxy_candidate *plist, struct FORWARD_LIST_LIST *fl)
{
	if (plist == NULL) {
		plist = xmalloc(sizeof(struct proxy_candidate));
		plist->prev = NULL;
	} else {
		while (plist->next != NULL)
			plist = plist->next;
		plist->next = xmalloc(sizeof(struct proxy_candidate));
		plist->next->prev = plist;
		plist = plist->next;
	}

	plist->next = NULL;
	plist->proxy = fl;

	return plist;
}

struct proxy_candidate *plist_remove(struct proxy_candidate *head, struct proxy_candidate *plist)
{
	struct proxy_candidate *ret;

	if (plist == head) {
		ret = head->next;
		plist->next->prev = NULL;

		xfree(plist);
	} else {
		ret = head;

		if (plist->next != NULL)
			plist->next->prev = plist->prev;
		plist->prev->next = plist->next;

		xfree(plist);
	}

	return ret;
}

void plist_free(struct proxy_candidate *plist)
{
	struct proxy_candidate *tmp;

	while (plist != NULL) {
		tmp = plist->next;
		xfree(plist);
		plist = tmp;
	}
}

struct FORWARD_LIST_LIST *proxy_select(CONNECTION * connection, struct proxy_candidate *plist)
{
	int icpcount = 0, fd = -1, x, timeout;
	int parenthitcount = 0, parentmisscount = 0, siblingcount = 0, normalcount = 0;
	int parenthitrand = -1, parentmissrand = -1, siblingrand = -1, normalrand = -1;
	char buf[1024];
	struct proxy_candidate *head = plist, *ptmp;
	struct icp_packet icppacket;
	struct sockaddr_in saddr;
	fd_set fdlist;
	struct timeval tv;

	if (connection->header->type != HTTP_CONNECT) {
		/* squid's URL parsing is stupid.. it thinks the URL's
		   http://host:80/ and http://host/ are distinct and cache's them seperately;
		   therefore, the port is only sent when it's not using the standard protocol
		   port. Middleman doesn't have this problem and will accept either */
		if (protocol_port(connection->header->proto) == connection->header->port)
			snprintf(buf, sizeof(buf), "%s://%s%s", connection->header->proto, connection->header->host, connection->header->file);
		else
			snprintf(buf, sizeof(buf), "%s://%s:%d%s", connection->header->proto, connection->header->host, connection->header->port, connection->header->file);

		icppacket.opcode = ICP_OP_QUERY;
		icppacket.version = 3;
		icppacket.mlen = htons(ICP_PACKET_LEN + strlen(buf) + 5);
		icppacket.options = 0;
		icppacket.optiondata = 0;

		/* this proxy's address */
		icppacket.hostaddr = 0;
		if (connection->interface != NULL)
			inet_pton(AF_INET, connection->interface, (struct addr_in *)&icppacket.hostaddr);

		/* client's address */
		memset(icppacket.payload, 0, 4);
		if (connection->ip != NULL)
			inet_pton(AF_INET, connection->ip, (struct addr_in *)&icppacket.payload);

		s_strncpy(&icppacket.payload[4], buf, ICP_MESSAGE_MAX - 4);

		for (; plist; plist = plist->next) {
			if (plist->proxy->icptype != ICP_NONE) {
				plist->icppacket.opcode = 255;

				if (fd == -1) {
					fd = socket(AF_INET, SOCK_DGRAM, 0);
					if (fd == -1)
						continue;

					saddr.sin_family = AF_INET;
					saddr.sin_addr.s_addr = INADDR_ANY;
					saddr.sin_port = 0;

					x = bind(fd, (struct sockaddr *) &saddr, sizeof(struct sockaddr));
					if (x == -1) {
						close(fd);
						continue;
					}
				}

				plist->reqnum = rand() % ~0;
				icppacket.rnum = htonl(plist->reqnum);

				saddr.sin_family = AF_INET;
				inet_pton(AF_INET, plist->proxy->proxy, &saddr.sin_addr);
				saddr.sin_port = htons(plist->proxy->icpport);

				icp_packet_send(fd, &saddr, &icppacket);
				icpcount++;
			}
		}

		plist = head;

		timeout = atomic_read_mutex(&global->cache->lock, &global->cache->icptimeout);

		tv.tv_sec = timeout / 1000;
		tv.tv_usec = (timeout % 1000) * 1000;

		while (icpcount) {
			FD_ZERO(&fdlist);
			FD_SET(fd, &fdlist);

			x = select(fd + 1, &fdlist, NULL, NULL, &tv);
			if (x == 1) {
				x = icp_packet_receive(fd, &saddr, &icppacket);
				if (x != -1) {
					inet_ntop(AF_INET, &saddr.sin_addr, buf, sizeof(buf));
					putlog(MMLOG_ICP, "response opcode from %s:%d: %s", buf, ntohs(saddr.sin_port), icp_opcode_name(icppacket.opcode));

					for (; plist; plist = plist->next) {
						if (plist->proxy->icptype != ICP_NONE && plist->reqnum == icppacket.rnum) {
							memcpy(&plist->icppacket, &icppacket, icppacket.mlen);
							icpcount--;
							break;
						}
					}

					if (plist == NULL)
						putlog(MMLOG_ICP, "received unknown ICP response number");

					plist = head;
				} else
					putlog(MMLOG_ICP, "failed to receive packet");
			} else if (x == 0 || errno != EINTR)
				break;

		}

		if (fd != -1)
			close(fd);

	}

	for (; plist; plist = ptmp) {
		ptmp = plist->next;

		/* count the usable proxies, so we have a range to select a random number from */
		if (plist->proxy->icptype == ICP_SIBLING) {
			if (plist->icppacket.opcode == ICP_OP_HIT)
				siblingcount++;
			continue;
		} else if (plist->proxy->icptype == ICP_PARENT) {
			if (plist->icppacket.opcode == ICP_OP_HIT)
				parenthitcount++;
			else if (plist->icppacket.opcode == ICP_OP_MISS)
				parentmisscount++;
			continue;
		} else
			normalcount++;
	}

	plist = head;

	if (parenthitcount)
		parenthitrand = rand() % parenthitcount;
	if (parentmisscount)
		parentmissrand = rand() % parentmisscount;
	if (siblingcount)
		siblingrand = rand() % siblingcount;
	if (normalcount)
		normalrand = rand() % normalcount;

	for (; plist; plist = plist->next) {
		/* the peer selection algorithm works as follows:
		   fetch file from a random sibling if any have it, otherwise
		   fetch file from a random parent proxy that does have it, otherwise
		   fetch file from a random parent proxy that doesn't have it, otherwise
		   fetch from a normal proxy.
		*/

		if (plist->proxy->icptype != ICP_NONE && plist->icppacket.opcode == 255)
			putlog(MMLOG_ICP, "No response from %s:%d", plist->proxy->proxy, plist->proxy->icpport);
		else if (plist->proxy->icptype == ICP_SIBLING && plist->icppacket.opcode == ICP_OP_HIT && siblingrand-- == 0)
			return plist->proxy;
		else if (siblingcount == 0 && plist->proxy->icptype == ICP_PARENT) {
			if (plist->icppacket.opcode == ICP_OP_HIT && parenthitrand-- == 0)
				return plist->proxy;
			else if (parenthitcount == 0 && plist->icppacket.opcode == ICP_OP_MISS && parentmissrand-- == 0) 
				return plist->proxy;
		} else if (parenthitcount == 0 && parentmisscount == 0 && siblingcount == 0 && normalrand-- == 0)
			return plist->proxy;
	}

	ASSERT(head == NULL || (parenthitcount == 0 && parentmisscount == 0 && siblingcount == 0 && normalcount == 0));

	return NULL;
}

int icp_packet_send(int fd, struct sockaddr_in *dest, struct icp_packet *icppacket)
{
	return sendto(fd, icppacket, ntohs(icppacket->mlen), 0, (struct sockaddr *) dest, sizeof(struct sockaddr_in));
}

int icp_packet_receive(int fd, struct sockaddr_in *saddr, struct icp_packet *icppacket)
{
	int x, len = 0, gotlength = FALSE;
	socklen_t sl = sizeof(struct sockaddr_in);

	/* I know this may not always work... but since this algorithm is designed to select
	   the best ICP peer, receiving a fragmented packet can be an indication
	   that this peer isn't a good choice anyways */
	do {
		x = recvfrom(fd, (void *) icppacket + len, (gotlength == FALSE) ? ICP_MESSAGE_MAX - len : icppacket->mlen - len, MSG_DONTWAIT, (struct sockaddr *) saddr, &sl);
		if (x <= 0)
			return -1;

		if (gotlength == FALSE && x >= 4) {
			gotlength = TRUE;

			icppacket->mlen = ntohs(icppacket->mlen);

			/* there's no need to check if the given size is too large.. since it can't be due to the maximum
			   value of icppacket->mlen. */
			if (icppacket->mlen < ICP_PACKET_LEN) {
				putlog(MMLOG_ICP, "ICP packet too short");
				return -1;
			}
		}

		len += x;
	} while ((gotlength == FALSE && len < ICP_MESSAGE_MAX) || (gotlength == TRUE && len < icppacket->mlen));

	icppacket->rnum = ntohl(icppacket->rnum);

	return len;
}

char *icp_opcode_name(int opcode)
{
	switch (opcode) {
	case 0:
		return "ICP_OP_INVALID";
	case 1:
		return "ICP_OP_QUERY";
	case 2:
		return "ICP_OP_HIT";
	case 3:
		return "ICP_OP_MISS";
	case 4:
		return "ICP_OP_ERR";
	case 10:
		return "ICP_OP_SECHO";
	case 11:
		return "ICP_OP_DECHO";
	case 21:
		return "ICP_OP_MISS_NOFETCH";
	case 22:
		return "ICP_OP_DENIED";
	case 23:
		return "ICP_OP_HIT_OBJ";
	default:
		return "INVALID";
	}
}

void icp_thread() {
	int fd = -1, x, port;
	struct sockaddr_in saddr;
	struct icp_packet icppacket;
	fd_set fdlist;
	char *newurl;
	URL *url;

	port = atomic_read_mutex(&global->cache->lock, &global->cache->icpport);
	if (port == 0) return;

	fd = socket(AF_INET, SOCK_DGRAM, 0);
	if (fd == -1) goto error;

	saddr.sin_family = AF_INET;
	saddr.sin_addr.s_addr = INADDR_ANY;
	saddr.sin_port = htons(port);

	x = bind(fd, (struct sockaddr *)&saddr, sizeof(struct sockaddr_in));
	if (x == -1) goto error;

	while (1) {
		FD_ZERO(&fdlist);
		FD_SET(fd, &fdlist);

		x = select(fd + 1, &fdlist, NULL, NULL, NULL);
		if (x == 1) {
			x = icp_packet_receive(fd, &saddr, &icppacket);
			if (x != -1) {
				icppacket.rnum = htonl(icppacket.rnum);

				if (icppacket.opcode == ICP_OP_QUERY) {
					if (icppacket.mlen <= 24 || memchr(icppacket.payload + 4, '\0', icppacket.mlen - ICP_PACKET_LEN - 4) == NULL) {
						icppacket.opcode = ICP_OP_INVALID;
						icppacket.mlen = htons(ICP_PACKET_LEN);
						icp_packet_send(fd, &saddr, &icppacket);
						putlog(MMLOG_ICP, "received ICP QUERY with no payload");
						continue;
					}

					url = url_parse(icppacket.payload + 4);
					if (url == NULL) {
						icppacket.opcode = ICP_OP_INVALID;
						icppacket.mlen = htons(ICP_PACKET_LEN);
						icp_packet_send(fd, &saddr, &icppacket);
						putlog(MMLOG_ICP, "received invalid URL in query");
						continue;
					}

					/* get rid of the client host in the response payload 
					   (probably unnecessary) */
					memmove(icppacket.payload, icppacket.payload + 4, icppacket.mlen - ICP_PACKET_LEN - 4);
					icppacket.mlen = htons(icppacket.mlen - 4);

					newurl = url_create(url);
					if (cache_exists(global->cache, newurl)) {
						putlog(MMLOG_ICP, "hit: %s", newurl);
						icppacket.opcode = ICP_OP_HIT;
					} else {
						putlog(MMLOG_ICP, "miss: %s", newurl);
						icppacket.opcode = ICP_OP_MISS;
					}
					xfree(newurl);

					icp_packet_send(fd, &saddr, &icppacket);
				} else {
					putlog(MMLOG_ICP, "unhandled ICP opcode: %s", icp_opcode_name(icppacket.opcode));
					icppacket.opcode = ICP_OP_INVALID;
					icppacket.mlen = htons(ICP_PACKET_LEN);

					icp_packet_send(fd, &saddr, &icppacket);
				}
			}
		}
	}

	error:
	if (fd != -1) close(fd);

	putlog(MMLOG_ERROR, "failed to start ICP thread");
}
