/*
    MiddleMan filtering proxy server
    Copyright (C) 2002  Jason McLaughlin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <signal.h>
#include <fcntl.h>
#include <pwd.h>
#include <grp.h>
#include <time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include "proto.h"

#ifdef USE_SYSLOG
#include <syslog.h>
#endif				/* USE_SYSLOG */

int loglevel = MMLOG_DEBUG - 1;
char configfile[256] = "", logfile[256] = "", username[256] = "", group[256] = "";
TEMPLATES *templates = NULL;
ACCESS_LIST *access_list = NULL;
HEADER_LIST *header_list = NULL;
PROFILES_LIST *profiles_list = NULL;
FILTER_LIST *filter_list = NULL;
COOKIE_LIST *cookie_list = NULL;
REWRITE_LIST *rewrite_list = NULL;
MIME_LIST *mime_list = NULL;
REDIRECT_LIST *redirect_list = NULL;
KEYWORD_LIST *keyword_list = NULL;
FORWARD_LIST *forward_list = NULL;
LIMITS_LIST *limits_list = NULL;
EXTERNAL *external = NULL;
THREADLIST threads[MAXTHREADS];
NETWORK *network = NULL;
GLOBAL *global = NULL;
int gmtoffset;

int main(int argc, char **argv)
{
	int fpid, x;
	char pidfile[256] = "";
	struct stat fileinfo;
	struct passwd *pwd = NULL;
	struct group *grp = NULL;

	srand(time(NULL) + getpid());

	if (argc < 2) {
		show_help(argv);
		exit(EXIT_SUCCESS);
	}

	while ((x = getopt(argc, argv, "hp:c:l:d:u:g:")) != EOF) {
		switch (x) {
		case 'c':
			s_strncpy(configfile, optarg, sizeof(configfile));
			break;
		case 'l':
			s_strncpy(logfile, optarg, sizeof(logfile));
			break;
		case 'p':
			s_strncpy(pidfile, optarg, sizeof(pidfile));
			break;
		case 'd':
			loglevel = atoi(optarg);
			break;
		case 'u':
			s_strncpy(username, optarg, sizeof(username));
			break;
		case 'g':
			s_strncpy(group, optarg, sizeof(group));
			break;
		case 'h':
			show_help(argv);
			exit(EXIT_SUCCESS);
		}
	}

	while (--argc > 0)
		memset(argv[argc], 0, strlen(argv[argc]));

	if (*configfile) {
		if (stat(configfile, &fileinfo) == -1) {
			fprintf(stderr, "couldn't stat %s\n", configfile);
			exit(EXIT_FAILURE);
		}
	} else {
		fprintf(stderr, "config file option missing\n");
		exit(EXIT_FAILURE);
	}

	if (strcmp(group, "")) {
		grp = getgrnam(group);
		if (grp == NULL) {
			fprintf(stderr, "getgrnam: unknown group\n");
			exit(EXIT_FAILURE);
		}

		x = setgid(grp->gr_gid);
		if (x == -1) {
			perror("setgid");
			exit(EXIT_FAILURE);
		}
	} else {
		grp = getgrgid(getgid());
		if (grp == NULL)
			snprintf(group, sizeof(group), "%d", getgid());
		else
			s_strncpy(group, grp->gr_name, sizeof(group));
	}

	if (strcmp(username, "")) {
		pwd = getpwnam(username);
		if (pwd == NULL) {
			fprintf(stderr, "getpwnam: unknown user\n");
			exit(EXIT_FAILURE);
		}

		x = setuid(pwd->pw_uid);
		if (x == -1) {
			perror("setuid");
			exit(EXIT_FAILURE);
		}
	} else {
		pwd = getpwuid(getuid());
		if (pwd == NULL)
			snprintf(username, sizeof(username), "%d", getuid());
		else
			s_strncpy(username, pwd->pw_name, sizeof(username));
	}

	fpid = fork();

	switch (fpid) {
	case -1:
		fprintf(stderr, "failed to fork daemon\n");
		exit(EXIT_FAILURE);
		break;
	case 0:
		setsid();

		if (*pidfile) {
			x = pid_check(pidfile);

			switch (x) {
			case 0:
				break;
			case -1:
				exit(EXIT_FAILURE);
			default:
				exit(EXIT_FAILURE);
			}
		}

		close(STDIN_FILENO);
		close(STDOUT_FILENO);
		close(STDERR_FILENO);

		mainloop();
	}

	exit(EXIT_SUCCESS);
}

void show_help(char **argv)
{
	fprintf(stderr, "MiddleMan filtering proxy server v%s (c)2002 Jason McLaughlin\n\n", VERSION);
	fprintf(stderr, "Usage: %s [options]\n\n", argv[0]);
	fprintf(stderr, " -c <file>  : location of config file\n");
#ifndef USE_SYSLOG
	fprintf(stderr, " -l <file>  : file to log actvity to\n");
#endif
	fprintf(stderr, " -p <file>  : PID file\n");
	fprintf(stderr, " -u <username> : run as alternate user\n");
	fprintf(stderr, " -g <groupname> : run in alternate group\n");
	fprintf(stderr, " -d <level> : set log level (default: %d)\n", loglevel);
	fprintf(stderr, " -h : help\n\n");
	fprintf(stderr, " Add any of the following to specify logging detail:\n");
	fprintf(stderr, " %7d = requests\n", MMLOG_REQUEST);
	fprintf(stderr, " %7d = network\n", MMLOG_NETWORK);
	fprintf(stderr, " %7d = url filtering\n", MMLOG_FILTER);
	fprintf(stderr, " %7d = header filtering\n", MMLOG_HEADER);
	fprintf(stderr, " %7d = mime filtering\n", MMLOG_MIME);
	fprintf(stderr, " %7d = cookie filtering\n", MMLOG_COOKIE);
	fprintf(stderr, " %7d = redirections\n", MMLOG_REDIRECT);
	fprintf(stderr, " %7d = templates\n", MMLOG_TEMPLATE);
	fprintf(stderr, " %7d = keyword filtering\n", MMLOG_KEYWORDS);
	fprintf(stderr, " %7d = rewriting\n", MMLOG_REWRITE);
	fprintf(stderr, " %7d = limits\n", MMLOG_LIMITS);
	fprintf(stderr, " %7d = caching\n", MMLOG_CACHE);
	fprintf(stderr, " %7d = prefetching\n", MMLOG_PREFETCH);
	fprintf(stderr, " %7d = ICP\n", MMLOG_ICP);
	fprintf(stderr, " %7d = forwarding\n", MMLOG_FORWARD);
	fprintf(stderr, " %7d = security\n", MMLOG_SECURITY);
	fprintf(stderr, " %7d = warnings\n", MMLOG_WARN);
	fprintf(stderr, " %7d = errors\n", MMLOG_ERROR);
	fprintf(stderr, " %7d = debug\n", MMLOG_DEBUG);
}

/*
check if pidfile exists and whether or not the pid inside is active, otherwise
write current pid.
*/
int pid_check(char *pidfile)
{
	int i = 0, x;
	FILE *fptr;

	fptr = fopen(pidfile, "r");
	if (fptr == NULL)
		goto makepidfile;

	x = fscanf(fptr, "%d", &i);
	if (x == 0) {
		fclose(fptr);
		goto makepidfile;
	}

	x = kill(i, SIGCHLD);
	if (x == 0) {
		fclose(fptr);
		return i;
	}

      makepidfile:
	unlink(pidfile);
	fptr = fopen(pidfile, "w");
	if (fptr == NULL)
		return TRUE;

	fprintf(fptr, "%u\n", (unsigned int) getpid());

	fclose(fptr);
	return FALSE;
}

/*
things that only need to be done once at startup
*/
void config()
{
	int i;
	time_t t, curtime = time(NULL);
	struct tm tt;

	gmtime_r(&curtime, &tt);
	t = mktime(&tt);

	/* tt.tm_isdst can also be -1 */
	gmtoffset = t - curtime - (tt.tm_isdst == 1) ? tt.tm_isdst : 0;

#ifdef USE_SYSLOG
	openlog("mman", LOG_PID, LOG_DAEMON);
#endif

	global = xmalloc(sizeof(GLOBAL));

	global->general = xmalloc(sizeof(GENERAL));
	pthread_rwlock_init(&global->general->lock, NULL);
	global->general->tempdir = xstrdup("/tmp");
	global->general->ctimeout = 10;
	global->general->timeout = 60;
	global->general->keeptimeout = 120;
	global->general->maxbuffer = 1000000;
	global->general->linelength = 180;
	global->general->poolsize = POOL_SIZE;
	global->general->pooltimeout = POOL_TIMEOUT;
	global->general->cportrange = NULL;
	global->general->compressout = global->general->compressin = FALSE;
	global->general->encodemime = NULL;
	global->general->emp = NULL;

	global->logbuffer = xmalloc(sizeof(LOGBUFFER));
	global->logbuffer->entries = 0;
	global->logbuffer->size = LOGBUFFERSIZE;
	global->logbuffer->head = global->logbuffer->tail = NULL;
	pthread_rwlock_init(&global->logbuffer->lock, NULL);

	global->ftp = xmalloc(sizeof(FTP));
	global->ftp->timeout = 15;
	global->ftp->pasv = FALSE;
	global->ftp->anonlogin = xstrdup(ANONLOGIN);
	global->ftp->anonpass = xstrdup(ANONPASS);
	pthread_rwlock_init(&global->ftp->lock, NULL);

	global->dnsbl = xmalloc(sizeof(DNSBL));
	global->dnsbl->enabled = FALSE;
	global->dnsbl->template = NULL;
	global->dnsbl->domain = NULL;
	global->dnsbl->badips = NULL;
	pthread_rwlock_init(&global->dnsbl->lock, NULL);

	pcre_free = xfree;
	pcre_malloc = (void *) xmalloc;

	global->dns_cache = hash_create(DNS_HASH_SIZE);
	pthread_mutex_init(&global->dns_cache_lock, NULL);

	pthread_mutex_init(&global->unlink_lock, NULL);
	pthread_cond_init(&global->unlink_cond, NULL);
	global->unlink_head = global->unlink_tail = NULL;

	global->cache = cache_init();
	global->pool = pool_init();
	//global->timers = timers_init();
	global->prefetch = prefetch_init();

	signal_setup();
	net_init();

	for (i = 0; i < MAXTHREADS; i++) {
		threads[i].flags = THREAD_UNUSED;
		pthread_mutex_init(&threads[i].lock, NULL);
		pthread_cond_init(&threads[i].cond, NULL);
		threads[i].arg = NULL;
	}

	config_load(3, configfile);

	prefetch_init_threads(global->prefetch);
	thread_create(icp_thread, NULL);
	thread_create(unlink_thread, NULL);

	putlog(MMLOG_DEBUG, "GMT offset is -%u seconds", gmtoffset);
}

int config_load(int overwrite, char *file)
{
	XML_LIST *xml_list;

	xml_list = xml_load(NULL, file);

	if (xml_list == NULL)
		return FALSE;

	global_load(xml_list);

	if (access_list != NULL) {
		pthread_rwlock_wrlock(&access_list->lock);

		if (overwrite) {
			access_list->id = 0;
			access_ll_free(access_list->allow);
			access_ll_free(access_list->deny);
			access_list->allow = access_list->deny = NULL;
		}

		access_load(access_list, xml_list);

		pthread_rwlock_unlock(&access_list->lock);
	} else
		access_list = access_load(NULL, xml_list);

	if (profiles_list != NULL) {
		pthread_rwlock_wrlock(&profiles_list->lock);

		if (overwrite) {
			profiles_list->id = 0;
			profiles_list_free(profiles_list->profiles);
			profiles_list->profiles = NULL;
		}

		profiles_load(profiles_list, xml_list);

		pthread_rwlock_unlock(&profiles_list->lock);
	} else
		profiles_list = profiles_load(NULL, xml_list);


	if (filter_list != NULL) {
		pthread_rwlock_wrlock(&filter_list->lock);

		if (overwrite) {
			filter_list->id = 0;
			filter_ll_free(filter_list->allow);
			filter_ll_free(filter_list->deny);
			filter_list->allow = filter_list->deny = NULL;
			FREE_AND_NULL(filter_list->dtemplate);
		}

		filter_load(filter_list, xml_list);

		pthread_rwlock_unlock(&filter_list->lock);
	} else
		filter_list = filter_load(NULL, xml_list);

	if (header_list != NULL) {
		pthread_rwlock_wrlock(&header_list->lock);

		if (overwrite) {
			header_list->id = 0;
			header_ll_free(header_list->allow);
			header_ll_free(header_list->deny);
			header_ll_free(header_list->insert);
			header_list->allow = header_list->deny = header_list->insert = NULL;
		}

		header_load(header_list, xml_list);

		pthread_rwlock_unlock(&header_list->lock);
	} else
		header_list = header_load(NULL, xml_list);

	if (cookie_list != NULL) {
		pthread_rwlock_wrlock(&cookie_list->lock);

		if (overwrite) {
			cookie_list->id = 0;
			cookie_ll_free(cookie_list->allow);
			cookie_ll_free(cookie_list->deny);
			cookie_list->allow = cookie_list->deny = NULL;
		}

		cookie_load(cookie_list, xml_list);

		pthread_rwlock_unlock(&cookie_list->lock);
	} else
		cookie_list = cookie_load(NULL, xml_list);

	if (rewrite_list != NULL) {
		pthread_rwlock_wrlock(&rewrite_list->lock);

		if (overwrite) {
			rewrite_list->id = 0;
			rewrite_list_free(rewrite_list->rewrite);
			rewrite_list->rewrite = NULL;
		}

		rewrite_load(rewrite_list, xml_list);

		pthread_rwlock_unlock(&rewrite_list->lock);
	} else
		rewrite_list = rewrite_load(NULL, xml_list);

	if (mime_list != NULL) {
		pthread_rwlock_wrlock(&mime_list->lock);

		if (overwrite) {
			mime_list->id = 0;
			mime_ll_free(mime_list->allow);
			mime_ll_free(mime_list->deny);
			mime_list->allow = mime_list->deny = NULL;
			FREE_AND_NULL(mime_list->dtemplate);
		}

		mime_load(mime_list, xml_list);

		pthread_rwlock_unlock(&mime_list->lock);
	} else
		mime_list = mime_load(NULL, xml_list);

	if (redirect_list != NULL) {
		pthread_rwlock_wrlock(&redirect_list->lock);

		if (overwrite) {
			redirect_list->id = 0;
			redirect_list_free(redirect_list->redirect_list);
			redirect_list->redirect_list = NULL;
		}

		redirect_load(redirect_list, xml_list);

		pthread_rwlock_unlock(&redirect_list->lock);
	} else
		redirect_list = redirect_load(NULL, xml_list);

	if (keyword_list != NULL) {
		pthread_rwlock_wrlock(&keyword_list->lock);

		if (overwrite) {
			keyword_list->id = 0;
			keyword_list_free(keyword_list->keyword_list);
			keyword_list->keyword_list = NULL;
		}

		keyword_load(keyword_list, xml_list);

		pthread_rwlock_unlock(&keyword_list->lock);
	} else
		keyword_list = keyword_load(NULL, xml_list);

	if (forward_list != NULL) {
		pthread_rwlock_wrlock(&forward_list->lock);

		if (overwrite) {
			forward_list->id = 0;
			forward_list_free(forward_list->forward_list);
			forward_list->forward_list = NULL;
		}

		forward_load(forward_list, xml_list);

		pthread_rwlock_unlock(&forward_list->lock);
	} else
		forward_list = forward_load(NULL, xml_list);

	if (limits_list != NULL) {
		pthread_rwlock_wrlock(&limits_list->lock);

		if (overwrite) {
			limits_list->id = 0;
			limits_ll_free(limits_list->limits_list);
			limits_list->limits_list = NULL;
		}

		limits_load(limits_list, xml_list);

		pthread_rwlock_unlock(&limits_list->lock);
	} else
		limits_list = limits_load(NULL, xml_list);

	if (templates != NULL) {
		pthread_rwlock_wrlock(&templates->lock);

		if (overwrite) {
			templates->id = 0;
			templates_list_free(templates->template_list);
			templates->template_list = NULL;
			FREE_AND_NULL(templates->path);
		}

		templates_load(templates, xml_list);

		pthread_rwlock_unlock(&templates->lock);
	} else
		templates = templates_load(NULL, xml_list);

	if (external != NULL) {
		pthread_rwlock_wrlock(&external->lock);

		if (overwrite) {
			external->id = 0;
			external_list_free(external->external_list);
			external->external_list = NULL;
		}

		external_load(external, xml_list);

		pthread_rwlock_unlock(&external->lock);
	} else
		external = external_load(NULL, xml_list);

	pthread_mutex_lock(&global->cache->lock);
	if (global->cache->stores != NULL) {
		if (overwrite) {
			global->cache->id = 0;
			stores_list_free(global->cache->stores);
			global->cache->stores = NULL;
		}

		global->cache->stores = cache_load(global->cache->stores, xml_list);
	} else global->cache->stores = cache_load(NULL, xml_list);

	stores_update(global->cache);

	pthread_mutex_unlock(&global->cache->lock);

	/* only do this at startup */
	if (overwrite == 3) {
		if (network != NULL) {
			pthread_rwlock_wrlock(&network->lock);

			network = network_load(network, xml_list);
			network_check(network);

			pthread_rwlock_unlock(&network->lock);
		} else {
			network = network_load(NULL, xml_list);
			network_check(network);
		}
	}

	xml_list_free(xml_list);

	return TRUE;
}

/* 
main event loop; accept connections, check access list, then create thread to
continue connection.
*/
void mainloop()
{
	int x;
	CONNECTION *connection;
	struct ACCESS_LIST_LIST *al;

	config();

	while (1) {
		connection = net_accept(1);

		//timers_check(global->timers);
		limits_reset(limits_list);

		if (connection != NULL) {
			al = access_check(access_list, connection, NULL, NULL);
			if (al != NULL) {
				x = process_new(connection);
				if (x != 0) {
					putlog(MMLOG_ERROR, "failed to create thread for %s", connection->ip);

					net_close(connection);
				}
			} else {
				putlog(MMLOG_SECURITY, "refused connect from %s on port %d", connection->ip, connection->port);

				net_close(connection);
			}

			pthread_rwlock_unlock(&access_list->lock);
		}

		pool_clean(global->pool);
	}
}

/*
create new thread
*/
int process_new(CONNECTION * connection)
{
	int thread = -1, i;

	for (i = 0; i < MAXTHREADS; i++) {
		pthread_mutex_lock(&threads[i].lock);
		if (threads[i].flags == THREAD_WAITING) {
			threads[i].flags = THREAD_IDLE;

			connection->thread = i;

			threads[i].arg = connection;
			pthread_cond_signal(&threads[i].cond);

			pthread_mutex_unlock(&threads[i].lock);

			return 0;
		}

		if (thread == -1 && threads[i].flags == THREAD_UNUSED)
			thread = i;

		pthread_mutex_unlock(&threads[i].lock);
	}

	if (thread == -1)
		return -1;
	else {
		threads[thread].flags = THREAD_IDLE;
		threads[thread].host = NULL;
		threads[thread].file = NULL;
		threads[thread].method = NULL;
		threads[thread].port = 0;
		threads[thread].requests = 0;

		threads[thread].ip = xstrdup(connection->ip);
	}

	connection->thread = thread;

	return thread_create(process_entry, connection);
}

int thread_create(void *func, void *arg)
{
	int perr;
	pthread_t thread_id;
	pthread_attr_t thread_attr;

	pthread_attr_init(&thread_attr);
	pthread_attr_setstacksize(&thread_attr, 1024 * 1024);
	pthread_attr_setdetachstate(&thread_attr, PTHREAD_CREATE_DETACHED);
	if (getuid() == 0)
		pthread_attr_setschedpolicy(&thread_attr, SCHED_FIFO);

	perr = pthread_create(&thread_id, &thread_attr, func, arg);

	pthread_attr_destroy(&thread_attr);

	return perr;
}


/*
entry function for new threads
*/
void process_entry(CONNECTION * connection)
{
	int x, ret;
	struct FILTER_LIST_LIST *filter_match;
	struct ACCESS_LIST_LIST *al;
	struct LIMITS_LIST_LIST *ll;
	char *headbuf = NULL, *ptr, buf[8192];
	FILEBUF *filebuf;
	HEADER *header;
	HOSTENT *hostent;
	struct timespec at;

      newconnection:
	/* write log message here so the pid matches */
	putlog(MMLOG_NETWORK, "allowed connect from %s on port %d", connection->ip, connection->port);

	pthread_mutex_lock(&threads[connection->thread].lock);
	threads[connection->thread].pid = (unsigned int) getpid();
	threads[connection->thread].flags = THREAD_BUSY;
	FREE_AND_STRDUP(threads[connection->thread].ip, connection->ip);
	pthread_mutex_unlock(&threads[connection->thread].lock);

	while (1) {
		/* this is done to reset profile list */
		al = access_check(access_list, connection, NULL, NULL);
		if (al == NULL) {
			pthread_rwlock_unlock(&access_list->lock);
			break;
		}
		access_setup(connection, al);
		pthread_rwlock_unlock(&access_list->lock);

		headbuf = header_get(connection, CLIENT, (connection->request) ? atomic_read_rwlock(&global->general->lock, &global->general->keeptimeout) : atomic_read_rwlock(&global->general->lock, &global->general->timeout));
		if (headbuf == NULL) {
			if (!connection->request) {
				putlog(MMLOG_WARN, "timeout waiting for header from %s", connection->ip);

				template_send(templates, "badrequest", connection, 400);
			}

			break;
		}

		/* this needs to be parsed twice; the first time to get the host/file, and the second
		   to let the rewrite rules take effect */
		connection->header = http_header_parse_request(headbuf);
		if (connection->header == NULL) {
			if (!connection->request) {
				putlog(MMLOG_WARN, "invalid header reveived from %s", connection->ip);

				template_send(templates, "badrequest", connection, 400);
			}

			xfree(headbuf);

			break;
		}

		/* this is done here to make it possible to have URL commands that
		   enable certain profile entries, which then may activate entries
 		   that modify the header. */
		if (url_command_extract(connection) == FALSE) {
			http_header_free(connection->header);
			xfree(headbuf);
			break;
		}

		bypass_check(connection);
		profiles_update(profiles_list, connection);
		http_header_free(connection->header);

		/* pass the client header through the rewrite rules before parsing */
		filebuf = filebuf_new();
		filebuf->data = headbuf;
		filebuf->size = strlen(headbuf) + 1;

		rewrite_do(rewrite_list, connection, filebuf, REWRITE_CLIENT, TRUE);

		headbuf = filebuf->data;
		filebuf->data = NULL;

		filebuf_free(filebuf);

		/* now parse the (possibly) rewritten header */
		connection->header = http_header_parse_request(headbuf);
		xfree(headbuf);
		if (connection->header == NULL) {
			if (!connection->request) {
				putlog(MMLOG_WARN, "invalid header reveived from %s", connection->ip);

				template_send(templates, "badrequest", connection, 400);
			}

			break;
		}

		/* this is done again to strip URL command from request */
		if (url_command_extract(connection) == FALSE) {
			http_header_free(connection->header);
			break;
		}

		/* determine if the connection should be kept alive with the information gathered
		   so far, this is incase a template or web interface is used */
		if (connection->header->type == HTTP_CONNECT)
			connection->keepalive_client = FALSE;
		else if (connection->header->keepalive == FALSE || connection->header->proxy_keepalive == FALSE)
			connection->keepalive_client = FALSE;
		else if (connection->header->proxy_keepalive == TRUE || connection->header->keepalive == TRUE)
			connection->keepalive_client = TRUE;
		else if (connection->header->version == HTTP_HTTP11)
			connection->keepalive_client = TRUE;
		else
			connection->keepalive_client = FALSE;

		if (connection->authenticate == TRUE && connection->header->proxy_authorization == NULL) {
			header = header_new();
			header->type = HTTP_RESP;
			header->code = 407;
			header->content_length = 0;
			header->proxy_authenticate = xstrdup("Basic");

			header_send(header, connection, CLIENT, HEADER_RESP);

			http_header_free(header);

			goto skip;
		} else if (connection->header->proxy_authorization != NULL) {
			ptr = strchr(connection->header->proxy_authorization, ' ');
			if (ptr != NULL) {
				ptr++;

				/* buf is large enough to hold the maximum possible string
				   length since a header line can only be 4kb */
				ret = from64tobits(buf, ptr);
				buf[ret] = '\0';

				ptr = strchr(buf, ':');
				if (ptr != NULL) {
					*ptr++ = '\0';

					if ((al = access_check(access_list, connection, buf, ptr))) {
						if (al->pamauth == TRUE) {
#ifdef HAVE_PAM
							ret = auth_pam(buf, ptr);
#else
							ret = FALSE;
#endif				/* HAVE_PAM */
							if (ret == TRUE) {
								putlog(MMLOG_SECURITY, "PAM authentication succeeded for %s", buf);

								access_setup(connection, al);
								connection->authenticate = FALSE;
							} else
								putlog(MMLOG_SECURITY, "PAM authentication failed for %s", buf);
						} else {
							putlog(MMLOG_SECURITY, "authentication succeeded for %s", buf);

							access_setup(connection, al);
							connection->authenticate = FALSE;
						}
					} else {
						putlog(MMLOG_SECURITY, "authentication failed for %s", buf);

						connection->authenticate = TRUE;
					}

					pthread_rwlock_unlock(&access_list->lock);
				}
			}

			if (connection->authenticate == TRUE) {
				header = header_new();
				header->type = HTTP_RESP;
				header->code = 407;
				header->content_length = 0;
				header->proxy_authenticate = xstrdup("Basic");

				header_send(header, connection, CLIENT, HEADER_RESP);

				http_header_free(header);

				goto skip;
			}
		}

		if (url_command_find(connection->url_command, "profiles")) {
			show_profiles(connection);

			goto skip;
		}

		/* redirect the request if any matching rules found (redirect_do will fill in the host, file and port
		   members of the connection struct if any are found) */
		x = redirect_do(redirect_list, connection, REDIRECT_REQUEST);
		if (x == TRUE) {
			/* 302 redirect sent, no need to continue */
			goto skip;
		}

		pthread_mutex_lock(&threads[connection->thread].lock);
		FREE_AND_STRDUP(threads[connection->thread].host, connection->header->host);
		FREE_AND_STRDUP(threads[connection->thread].file, connection->header->file);
		FREE_AND_STRDUP(threads[connection->thread].method, connection->header->method);
		threads[connection->thread].port = connection->header->port;
		pthread_mutex_unlock(&threads[connection->thread].lock);

		connection->header->header_filtered = header_filter(header_list, connection, HEADER_CLIENT);

		if (connection->header->type != HTTP_CONNECT) {
			if ((connection->header->type != HTTP_REQUEST && !strcasecmp(connection->header->host, INTERFACEURL)) || (connection->header->type == HTTP_REQUEST && !strncasecmp(&connection->header->file[1], INTERFACEURL, strlen(INTERFACEURL)))) {
				/* request for web interface */
				putlog(MMLOG_REQUEST, "request for web interface from %s", connection->ip);

				interface_handle_request(connection);

				goto skip;
			}

			if ((connection->header->type == HTTP_REQUEST && !(connection->access & ACCESS_HTTP)) || (connection->header->type == HTTP_PROXY && !(connection->access & ACCESS_PROXY))) {
				template_send(templates, "noaccess", connection, 404);

				goto skip;
			}
		} else if (!(connection->access & ACCESS_CONNECT)) {
			template_send(templates, "noaccess", connection, 404);

			goto skip;
		} else {
			pthread_rwlock_rdlock(&global->general->lock);

			ret = TRUE;
			if (global->general->cportrange != NULL)
				ret = range_check(global->general->cportrange, connection->header->port);

			pthread_rwlock_unlock(&global->general->lock);

			if (ret == FALSE) {
				putlog(MMLOG_SECURITY, "denied CONNECT request to port %d", connection->header->port);

				template_send(templates, "noaccess", connection, 404);

				goto skip;
			}
		}

		if (connection->header->type == HTTP_REQUEST && connection->header->host == NULL) {
			/* this is a regular HTTP request */
			if ((connection->access & ACCESS_TRANSPARENT) && connection->header->host_header != NULL) {
				/* use Host: header if it's there */
				ptr = strchr(connection->header->host_header, ':');
				connection->header->host = xstrndup(connection->header->host_header, (ptr != NULL) ? ptr - connection->header->host_header : strlen(connection->header->host_header));
				if (ptr != NULL)
					connection->header->port = atoi(&ptr[1]);

				/* try redirecting again now that we've retrieved the host from the Host: header */
				x = redirect_do(redirect_list, connection, REDIRECT_REQUEST);
				if (x == TRUE) {
					/* 302 redirect sent */
					goto skip;
				}

				/* this feature causes a recursion where the proxy keeps
				   connecting to itself if an HTTP request is made to the proxy
				   which doesn't match a redirect rule and isn't a request for the web interface.
				   (the Host: header will have the proxies own hostname/port) 
				   There's no reliable way to detect this except to forbid connections
				   to websites on the same port as the proxy
				 */
				if (connection->header->port == connection->port) {
					template_send(templates, "nofile", connection, 404);

					goto skip;
				}
			} else {
				/* not a request for web interface, no host header, and no matching redirect rule */
				template_send(templates, "nofile", connection, 404);
				goto skip;
			}
		}

		if (url_command_find(connection->url_command, "filter")) {
			filter_check_show(connection);

			goto skip;
		}

		if (url_command_find(connection->url_command, "cache")) {
			cache_show(connection);

			goto skip;
		}

		filter_match = filter_check(filter_list, connection);
		if (filter_match != NULL) {
			putlog(MMLOG_FILTER, "blocked %s%s", connection->header->host, connection->header->file);

			template_send(templates, (filter_match->template != NULL) ? filter_match->template : (filter_list->dtemplate != NULL) ? filter_list->dtemplate : "blocked", connection, (connection->header->type == HTTP_CONNECT) ? 404 : 200);

			pthread_rwlock_unlock(&filter_list->lock);

			goto skip;
		}
		pthread_rwlock_unlock(&filter_list->lock);

		connection->transferred = 0;
		connection->transferlimit = ~0;

		ll = limits_check(limits_list, connection);
		if (ll != NULL) {
			if (connection->transferlimit == 0 || ll->action == POLICY_DENY) {
				/* blocked due to time restriction or deny action */
				putlog(MMLOG_LIMITS, "blocked due to time limit restrictions");

				template_send(templates, (ll->template != NULL) ? ll->template : "blocked", connection, 404);

				pthread_rwlock_unlock(&limits_list->lock);

				goto skip;
			}

			if (ll->maxrequests != 0 && ll->n_requests >= ll->maxrequests) {
				putlog(MMLOG_LIMITS, "blocked due to request limit restrictions");

				template_send(templates, (ll->template != NULL) ? ll->template : "maxrequests", connection, 404);

				pthread_rwlock_unlock(&limits_list->lock);

				goto skip;
			} else if (ll->maxbytes != 0)
				connection->transferlimit = (ll->maxbytes < ll->n_bytes) ? 0 : ll->maxbytes - ll->n_bytes;

			putlog(MMLOG_DEBUG, "connection transfer limit: %u", connection->transferlimit);

			if (ll->flags & LIMIT_CACHE)
				connection->flags |= CONNECTION_LIMITCACHE;
		}

		pthread_rwlock_unlock(&limits_list->lock);

		pthread_rwlock_rdlock(&global->dnsbl->lock);
		if (global->dnsbl->enabled == TRUE && !(connection->bypass & FEATURE_DNSBL) && global->dnsbl->domain != NULL && global->dnsbl->badips != NULL) {
			ptr = strchr(connection->header->host, '\0');
			for (x = 0; x < 2 && ptr != connection->header->host; ptr--)
				if (*ptr == '.')
					x++;

			if (ptr != connection->header->host)
				ptr += 2;

			snprintf(buf, sizeof(buf), "%s.%s", ptr, global->dnsbl->domain);
			putlog(MMLOG_DEBUG, "domain: %s", buf);
			hostent = net_dns(buf);
			if (hostent != NULL) {
				if (array_find(global->dnsbl->badips, buf)) {
					template_send(templates, (global->dnsbl->template != NULL) ? global->dnsbl->template : "blocked", connection, 404);

					pthread_rwlock_unlock(&global->dnsbl->lock);

					goto skip;
				}
			}
		}
		pthread_rwlock_unlock(&global->dnsbl->lock);


		if (url_command_find(connection->url_command, "offline"))
			connection->flags |= CONNECTION_OFFLINE;

		snprintf(buf, sizeof(buf), "%s://%s:%d%s", connection->header->proto, connection->header->host, connection->header->port, connection->header->file);

		if (url_command_find(connection->url_command, "fresh") == NULL && cacheable(connection->header)) {
			connection->cachemap = cache_open(connection, global->cache, buf, NULL, (connection->flags & CONNECTION_OFFLINE) ? CACHE_OFFLINE : 0);

			if (connection->cachemap != NULL) {
				putlog(MMLOG_CACHE, "hit: %s", connection->cachemap->key);

				/* validate if the client is requesting an uncached copy */
				if (!(connection->flags & CONNECTION_OFFLINE)) {
					cache_check(global->cache, connection->header);
					if (atomic_read_mutex(&global->cache->lock, &connection->cachemap->flags) & CACHE_VALIDATE) {
						connection->flags |= CONNECTION_VALIDATE;
						conditional_header(connection->header, connection->cachemap);
					}
				}
			}

			/* this must be done here, since cache_open is sometimes called from cache_check as well */
			pthread_mutex_lock(&global->cache->lock);
			if (connection->cachemap != NULL)
				global->cache->hit++;
			else
				global->cache->miss++;
			pthread_mutex_unlock(&global->cache->lock);
		}

		if (connection->flags & CONNECTION_OFFLINE) {
			if (connection->cachemap == NULL) {
				template_send(templates, "nocache", connection, 404);

				goto skip;
			}
		}
#ifdef HAVE_ZLIB
		if (atomic_read_rwlock(&global->general->lock, &global->general->compressin) == TRUE) {
			if (accepts_encoded_content(connection) == FALSE) {
				connection->flags |= CONNECTION_ENCODED;
				FREE_AND_STRDUP(connection->header->accept_encoding, ACCEPTED_ENCODINGS);
				putlog(MMLOG_DEBUG, "asking server for encoded content");
			}
		}
#endif				/* HAVE_ZLIB */

	      protocol_start:
		x = 0;

		if (connection->server == NULL && (connection->cachemap == NULL || (connection->flags & CONNECTION_VALIDATE))) {
			/* check if this request should be forwarded through another proxy */
			/* forward_do will fill in all the necessary members of the connection struct */
			connection->proxy_type = PROXY_DIRECT;
			connection->flags &= ~CONNECTION_LOGGEDIN;
			forward_do(forward_list, connection);

			x = protocol_start(connection);
		}

		if (x != -1) {
			if (connection->header->type != HTTP_CONNECT) {
				if (url_command_find(connection->url_command, "proxytest")) {
					show_proxytest(connection);

					goto skip;
				}

				/* only process files retrieved through HTTP */
				if (!strcasecmp(connection->header->proto, "http"))
					connection->flags |= CONNECTION_PROCESS;

				if (!strcasecmp(connection->header->proto, "http") || connection->proxy_type == PROXY_NORMAL || (connection->cachemap != NULL && !(connection->flags & CONNECTION_VALIDATE)))
					x = protocol_http(connection);
				else if (!strcasecmp(connection->header->proto, "ftp"))
					x = protocol_ftp(connection);
				else {
					template_send(templates, "badprotocol", connection, 501);

					goto skip;
				}
			} else
				x = protocol_connect(connection);
		}

		if (x < 0) {
			if (connection->cachemap != NULL) {
				/* fall back on cache file if connection fails */
				connection->flags &= ~CONNECTION_VALIDATE;

				putlog(MMLOG_CACHE, "using unvalidated cache for failed connection");

				goto protocol_start;
			}

			template_send(templates, "noconnect", connection, 503);

			putlog(MMLOG_HEADER, "error reading header from %s", connection->header->host);
		}

	      skip:

		if (connection->transferred > 0 && connection->transferlimit == 0)
			putlog(MMLOG_LIMITS, "transfer prematurely terminated due to transfer limit restrictions");

		limits_update(limits_list, connection);

		sock_flush(connection->client);

		if (connection->cachemap != NULL) {
			cache_close(global->cache, connection->cachemap);
			connection->cachemap = NULL;
		}

		if (connection->url_command != NULL) {
			url_command_free(connection->url_command);
			connection->url_command = NULL;
		}

		if (connection->htmlstream != NULL) {
			htmlstream_free(connection->htmlstream);
			connection->htmlstream = NULL;
		}

		connection->request++;
		connection->flags = 0;

		if (connection->server != NULL) {
			if (connection->keepalive_server == TRUE)
				pool_add(global->pool, connection->server, connection->header->proto, (connection->proxy_type == PROXY_NORMAL && connection->proxy_host != NULL) ? connection->proxy_host : connection->header->host, (connection->proxy_type == PROXY_NORMAL) ? connection->proxy_port : connection->header->port, connection->header->username, connection->header->password);
			else
				sock_close(connection->server);

			connection->server = NULL;
		}

		http_header_free(connection->header);

		putlog(MMLOG_DEBUG, "keepalive_client = %d", connection->keepalive_client);

		if (!connection->keepalive_client)
			break;

		pthread_mutex_lock(&threads[connection->thread].lock);
		FREE_AND_NULL(threads[connection->thread].host);
		FREE_AND_NULL(threads[connection->thread].file);
		FREE_AND_NULL(threads[connection->thread].method);
		threads[connection->thread].port = 0;
		threads[connection->thread].requests++;
		pthread_mutex_unlock(&threads[connection->thread].lock);

		connection->header = NULL;
		connection->rheader = NULL;

		FREE_AND_NULL(connection->proxy_host);
	}

	if (connection->server != NULL)
		sock_close(connection->server);

	putlog(MMLOG_NETWORK, "%s disconnected after making %d requests", connection->ip, connection->request);

	x = connection->thread;
	net_close(connection);

	pthread_mutex_lock(&threads[x].lock);
	threads[x].flags = THREAD_WAITING;
	FREE_AND_NULL(threads[x].ip);
	FREE_AND_NULL(threads[x].host);
	FREE_AND_NULL(threads[x].file);
	FREE_AND_NULL(threads[x].method);
	threads[x].port = 0;
	threads[x].requests = 0;

	/* wait for another connection to handle */
	at.tv_sec = time(NULL) + THREAD_TIMEOUT;
	at.tv_nsec = 0;

	ret = pthread_cond_timedwait(&threads[x].cond, &threads[x].lock, &at);
	if (ret == ETIMEDOUT || threads[x].arg == NULL) {
		threads[x].flags = THREAD_UNUSED;
		pthread_mutex_unlock(&threads[x].lock);

		pthread_exit(0);
	}

	putlog(MMLOG_DEBUG, "reusing thread");
	connection = threads[x].arg;
	threads[x].arg = NULL;
	pthread_mutex_unlock(&threads[x].lock);

	goto newconnection;
}

/*
save all config settings to a file
*/
int config_save(char *filename)
{
	int ret;
	XML_LIST *xml_list = NULL;

	/* reconstruct config file for all sections */
	xml_list = global_xml(xml_list);
	xml_list = network_xml(network, xml_list);
	xml_list = profiles_xml(profiles_list, xml_list);
	xml_list = templates_xml(templates, xml_list);
	xml_list = external_xml(external, xml_list);
	xml_list = access_xml(access_list, xml_list);
	xml_list = cache_xml(global->cache->stores, xml_list);
	xml_list = header_xml(header_list, xml_list);
	xml_list = cookie_xml(cookie_list, xml_list);
	xml_list = redirect_xml(redirect_list, xml_list);
	xml_list = keyword_xml(keyword_list, xml_list);
	xml_list = forward_xml(forward_list, xml_list);
	xml_list = limits_xml(limits_list, xml_list);
	xml_list = filter_xml(filter_list, xml_list);
	xml_list = mime_xml(mime_list, xml_list);
	xml_list = rewrite_xml(rewrite_list, xml_list);

	if (xml_list == NULL)
		return FALSE;

	while (xml_list->prev != NULL)
		xml_list = xml_list->prev;

	ret = xml_save(xml_list, filename);

	xml_list_free(xml_list);

	return ret;
}

void bypass_check(CONNECTION * connection)
{
	int x;
	char *ptr;
	struct url_command_t *url_command;

	if ((connection->access & ACCESS_BYPASS) && connection->url_command != NULL) {
		url_command = url_command_find(connection->url_command, "bypass");
		if (url_command != NULL) {
			ptr = url_command->options;
			if (ptr != NULL) {
				x = TRUE;

				for (; *ptr; ptr++) {
					switch (*ptr) {
					case '+':
						x = TRUE;
						break;
					case '-':
						x = FALSE;
						break;
					case 'f':
						if (x == TRUE)
							connection->bypass |= FEATURE_FILTER;
						else
							connection->bypass &= ~FEATURE_FILTER;
						break;
					case 'h':
						if (x == TRUE)
							connection->bypass |= FEATURE_HEADER;
						else
							connection->bypass &= ~FEATURE_HEADER;
						break;
					case 'm':
						if (x == TRUE)
							connection->bypass |= FEATURE_MIME;
						else
							connection->bypass &= ~FEATURE_MIME;
						break;
					case 'r':
						if (x == TRUE)
							connection->bypass |= FEATURE_REDIRECT;
						else
							connection->bypass &= ~FEATURE_REDIRECT;
						break;
					case 'c':
						if (x == TRUE)
							connection->bypass |= FEATURE_COOKIES;
						else
							connection->bypass &= ~FEATURE_COOKIES;
						break;
					case 'w':
						if (x == TRUE)
							connection->bypass |= FEATURE_REWRITE;
						else
							connection->bypass &= ~FEATURE_REWRITE;
						break;
					case 'e':
						if (x == TRUE)
							connection->bypass |= FEATURE_EXTERNAL;
						else
							connection->bypass &= ~FEATURE_EXTERNAL;
						break;
					case 'p':
						if (x == TRUE)
							connection->bypass |= FEATURE_FORWARD;
						else
							connection->bypass &= ~FEATURE_FORWARD;
						break;
					case 'k':
						if (x == TRUE)
							connection->bypass |= FEATURE_KEYWORDS;
						else
							connection->bypass &= ~FEATURE_KEYWORDS;
						break;
					case 'd':
						if (x == TRUE)
							connection->bypass |= FEATURE_DNSBL;
						else
							connection->bypass &= ~FEATURE_DNSBL;
						break;
					case 'l':
						if (x == TRUE)
							connection->bypass |= FEATURE_LIMITS;
						else
							connection->bypass &= ~FEATURE_LIMITS;
						break;
					}
				}
			} else
				connection->bypass = ~0;
		}
	}
}
