/*
    MiddleMan filtering proxy server
    Copyright (C) 2002  Jason McLaughlin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include "proto.h"

extern GLOBAL *global;
extern FORWARD_LIST *forward_list;

PREFETCH *prefetch_init() {
	PREFETCH *prefetch;

	prefetch = xmalloc(sizeof(PREFETCH));
	prefetch->enabled = FALSE;
	prefetch->queue = NULL;
	prefetch->threads = prefetch->queuesize = 0;
	prefetch->tags = NULL;

	prefetch->entries = 0;

	pthread_mutex_init(&prefetch->lock, NULL);
	pthread_cond_init(&prefetch->cond, NULL);

	return prefetch;
}

void prefetch_init_threads(PREFETCH *prefetch) {
	int threads;

	threads = atomic_read_mutex(&prefetch->lock, &prefetch->threads);

	for (; threads > 0; threads--)
		thread_create(prefetch_thread, prefetch);
}

void prefetch_thread(PREFETCH *prefetch) {
	struct PREFETCH_QUEUE *queue;

	while(1) {
		pthread_mutex_lock(&prefetch->lock);
		if (prefetch->queue == NULL) pthread_cond_wait(&prefetch->cond, &prefetch->lock);

		if (prefetch->queue != NULL) {
			queue = prefetch->queue;
			prefetch->queue = queue->next;
			prefetch->entries--;

			pthread_mutex_unlock(&prefetch->lock);

			protocol_prefetch(queue->url, queue->depth);

			prefetch_queue_free(queue);
		} else {
			/* this can happen if another thread steals the entry in the queue without
			   calling pthread_cond_wait */
			pthread_mutex_unlock(&prefetch->lock);
		}
	}
}

int protocol_prefetch(URL *url, int depth)
{
	int x;
	char buf[1024];
	CONNECTION *connection;

	snprintf(buf, sizeof(buf), "%s://%s:%d%s", url->proto, url->host, url->port, url->file);
	if (cache_exists(global->cache, buf) == TRUE) return FALSE;

	/* there's a bit of a race condition here... another thread can create the cache file
	   between this point and when we open it for writing... worst case scenario is we end
	   up downloading some of the file twice */

	connection = connection_new();

	connection->flags = CONNECTION_PREFETCH;
	connection->depth = depth;

	connection->header = header_new();
	connection->header->type = HTTP_PROXY;
	connection->header->version = HTTP_HTTP11;
	connection->header->method = xstrdup("GET");
	connection->header->proto = xstrdup(url->proto);
	connection->header->host = xstrdup(url->host);
	connection->header->file = xstrdup(url->file);
	connection->header->port = url->port;

	/* some websites actually care about this... */
	if (connection->header->port == HTTP_PORT)
		snprintf(buf, sizeof(buf), "%s://%s%s", connection->header->proto, connection->header->host, connection->header->file);
	else
		snprintf(buf, sizeof(buf), "%s://%s:%d%s", connection->header->proto, connection->header->host, connection->header->port, connection->header->file);
	connection->header->referer = xstrdup(buf);

#ifdef HAVE_ZLIB
	connection->header->accept_encoding = xstrdup(ACCEPTED_ENCODINGS);
#endif /* HAVE_ZLIB */

	connection->header->header = header_list_insert(connection->header->header, "User-Agent", PREFETCH_USERAGENT);

	if (url->username != NULL) connection->header->username = xstrdup(url->username);
	if (url->password != NULL) connection->header->password = xstrdup(url->password);

	forward_do(forward_list, connection);
	
	x = protocol_start(connection);

	if (x != -1) {
		if (!strcasecmp(url->proto, "http"))
			protocol_http(connection);
		else if (!strcasecmp(url->proto, "ftp"))
			protocol_ftp(connection);
	}

	if (connection->cachemap != NULL)
		cache_close(global->cache, connection->cachemap);

	if (connection->htmlstream != NULL)
		htmlstream_free(connection->htmlstream);

	if (connection->server != NULL) {
		if (connection->keepalive_server == TRUE)
			pool_add(global->pool, connection->server, connection->header->proto, (connection->proxy_type == PROXY_NORMAL && connection->proxy_host != NULL) ? connection->proxy_host : connection->header->host, (connection->proxy_type == PROXY_NORMAL) ? connection->proxy_port : connection->header->port, NULL, NULL);
		else
			sock_close(connection->server);
	}

	http_header_free(connection->header);
	connection_free(connection);

	return TRUE;
}

void prefetch_queue_add(PREFETCH *prefetch, URL *url, int depth)
{
	struct PREFETCH_QUEUE *queue, *tmp;

	pthread_mutex_lock(&prefetch->lock);

	if (prefetch->entries >= prefetch->queuesize)
		goto error;

	prefetch->entries++;

	queue = xmalloc(sizeof(struct PREFETCH_QUEUE));
	queue->url = url;
	queue->depth = depth;
	queue->next = NULL;
	

	if (prefetch->queue == NULL)
		prefetch->queue = queue;
	else {
		for (tmp = prefetch->queue; tmp->next; tmp = tmp->next);
		tmp->next = queue;
	}

	pthread_cond_signal(&prefetch->cond);
	pthread_mutex_unlock(&prefetch->lock);

	return;
error:
	url_free(url);

	pthread_mutex_unlock(&prefetch->lock);
}

void prefetch_queue_free(struct PREFETCH_QUEUE *queue) {
	url_free(queue->url);
	xfree(queue);
}

void prefetch_setup_callbacks(CONNECTION *connection, PREFETCH *prefetch) {
	char **tags, **pair;

	pthread_mutex_lock(&prefetch->lock);
	if (prefetch->enabled == FALSE) goto out;

	if (connection->rheader->content_type == NULL || strncasecmp(connection->rheader->content_type, "text/html", 9))
		goto out;

	connection->htmlstream = htmlstream_new();

	tags = prefetch->tags;

	for (; tags && *tags; tags++) {
		pair = string_break(*tags, ':');
		if (pair[1] != NULL)
			htmlstream_callback_add(connection->htmlstream, pair[0], prefetch_callback, connection);

		array_free(pair);
	}
		
out:
	pthread_mutex_unlock(&prefetch->lock);
}

void prefetch_callback(HTMLSTREAM *hs, struct htmlstream_node *node, CONNECTION *connection) {
	int i;
	URL *url = NULL;
	char **tags, **tagopts, *ptr;
	struct htmlstream_tag_property *property;

	pthread_mutex_lock(&global->prefetch->lock);
	if (global->prefetch->tags == NULL) goto out;

	tags = global->prefetch->tags;

	for (; tags && *tags; tags++) {
		tagopts = string_break(*tags, ':');
		if (tagopts[1] != NULL) {
			if (!strcasecmp(tagopts[0], node->tag->name)) {
				if (tagopts[2] != NULL) {
					i = atoi(tagopts[2]);
					if (i != 0 && connection->depth >= i) goto endloop;
				} else if (connection->depth >= 1) goto endloop;

				for (property = node->tag->properties; property; property = property->next) {
					if (!strcasecmp(property->name, tagopts[1])) {
						if (strchr(property->value, ':') != NULL)
							url = url_parse(property->value);
						else {
							/* relative URL */
							url = xmalloc(sizeof(URL));
							url->username = url->password = NULL;
							url->proto = xstrdup(connection->header->proto);
							url->host = xstrdup(connection->header->host);
							url->port = connection->header->port;
							if (property->value[0] == '/')
								ptr = xstrdup(property->value);
							else {
								ptr = xstrndup(connection->header->file, strrchr(connection->header->file, '/') - connection->header->file + 1);
								ptr = string_append(ptr, property->value);
							}
							url->file = url_path_fix(ptr);
							xfree(ptr);
						}
					}
				}
			}
		}

	endloop:
		array_free(tagopts);
		if (url != NULL) break;
	}

	out:
	pthread_mutex_unlock(&global->prefetch->lock);

	if (url != NULL)
		prefetch_queue_add(global->prefetch, url, connection->depth + 1);
}
