/*
    MiddleMan filtering proxy server
    Copyright (C) 2002  Jason McLaughlin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "../libntlm/ntlm.h"
#include "proto.h"

extern TEMPLATES *templates;
extern REWRITE_LIST *rewrite_list;
extern HEADER_LIST *header_list;
extern EXTERNAL *external;
extern MIME_LIST *mime_list;
extern REDIRECT_LIST *redirect_list;
extern KEYWORD_LIST *keyword_list;
extern GLOBAL *global;

int protocol_start(CONNECTION * connection)
{
	int proxyfd, x;
	SOCKET *sock;

	connection->keepalive_server = FALSE;

	if (connection->header->type != HTTP_CONNECT) {
		sock = pool_find(global->pool, connection->header->proto, (connection->proxy_type == PROXY_NORMAL && connection->proxy_host != NULL) ? connection->proxy_host : connection->header->host, (connection->proxy_type == PROXY_NORMAL) ? connection->proxy_port : connection->header->port, connection->header->username, connection->header->password);
		if (sock != NULL) {
			connection->server = sock;
			connection->keepalive_server = TRUE;
			connection->flags |= CONNECTION_LOGGEDIN;

			return sock->fd;
		}
	}

	switch (connection->proxy_type) {
	case PROXY_NORMAL:
		proxyfd = net_connect(connection->proxy_host, connection->proxy_port, atomic_read_rwlock(&global->general->lock, &global->general->ctimeout));
		if (proxyfd < 0)
			putlog(MMLOG_NETWORK, "connection to proxy server failed");
		else
			connection->server = sock_new(proxyfd);

		break;
	case PROXY_SOCKS4:
		proxyfd = net_connect(connection->proxy_host, connection->proxy_port, atomic_read_rwlock(&global->general->lock, &global->general->ctimeout));

		if (proxyfd >= 0) {
			connection->server = sock_new(proxyfd);

			x = proxy_socks4(connection);
			if (x < 0) {
				putlog(MMLOG_NETWORK, "connection to socks4 server failed");

				sock_close(connection->server);
				connection->server = NULL;

				return -1;
			}
		}

		break;
	case PROXY_CONNECT:
		proxyfd = net_connect(connection->proxy_host, connection->proxy_port, atomic_read_rwlock(&global->general->lock, &global->general->ctimeout));
		if (proxyfd >= 0) {
			connection->server = sock_new(proxyfd);

			x = proxy_connect(connection);
			if (x < 0) {
				putlog(MMLOG_NETWORK, "connection to proxy failed");

				sock_close(connection->server);
				connection->server = NULL;

				return -1;
			}

		}

		break;
	default:
		proxyfd = net_connect(connection->header->host, connection->header->port, atomic_read_rwlock(&global->general->lock, &global->general->ctimeout));
		if (proxyfd >= 0)
			connection->server = sock_new(proxyfd);

		break;
	}

	if (proxyfd < 0) {
		connection->server = NULL;

		return -1;
	}

	return proxyfd;
}

int protocol_reconnect(CONNECTION * connection)
{
	sock_close(connection->server);
	if (connection->client != NULL)
		sock_flush(connection->client);

	return protocol_start(connection);
}

/*
This function handles HTTP requests
*/
int protocol_http(CONNECTION * connection)
{
	int ret, filebuf_size = 0, content_length, chunked;
	char *headbuf, buf[4096], *ptr;
	FILEBUF *filebuf, *filebuf2, *postdata = NULL;
	struct EXTERNAL_LIST_LIST *external_list = NULL;
	struct MIME_LIST_LIST *mime_list_list;
	time_t t = 0;
	struct tm tt;

	if (connection->cachemap == NULL || (connection->flags & CONNECTION_VALIDATE)) {
		if (connection->header->content_length != -1) {
			/* the browser wants to send something to the remote site after the header (POST, PUT, etc.) */
			putlog(MMLOG_DEBUG, "post length: %d", connection->header->content_length);

			/* this needs to be buffered.. otherwise we won't have any POST data
			   after authentication */
			postdata = http_transfer_filebuf(connection, CLIENT, -1);
			rewrite_do(rewrite_list, connection, postdata, REWRITE_POST, TRUE);

			connection->header->chunked = FALSE;
			connection->header->content_length = postdata->size;
		}

		header_send(connection->header, connection, SERVER, (connection->proxy_type == PROXY_NORMAL) ? HEADER_FORWARD : HEADER_DIRECT);

		/* I hate abusing GOTO's like this... but I'm lazy :) */
	      auth_finished:
		if (postdata != NULL)
			net_filebuf_send(postdata, connection, SERVER);

	      reget_header:
		headbuf = header_get(connection, SERVER, atomic_read_rwlock(&global->general->lock, &global->general->timeout));
		if (headbuf == NULL) {
			if (!connection->keepalive_server && !(connection->flags & CONNECTION_AUTHENTICATED)) {
				template_send(templates, "noconnect", connection, 503);

				if (postdata != NULL)
					filebuf_free(postdata);

				return -1;
			} else {
				do {
					ret = protocol_reconnect(connection);

					if (ret == -1) {
						if (postdata != NULL)
							filebuf_free(postdata);

						return -1;
					}

					header_send(connection->header, connection, SERVER, (connection->proxy_type == PROXY_NORMAL) ? HEADER_FORWARD : HEADER_DIRECT);
					if (postdata != NULL)
						net_filebuf_send(postdata, connection, SERVER);

					headbuf = header_get(connection, SERVER, atomic_read_rwlock(&global->general->lock, &global->general->timeout));
				} while (headbuf == NULL && connection->keepalive_server == TRUE);

				if (headbuf == NULL) {
					if (postdata != NULL)
						filebuf_free(postdata);

					return -1;
				}
			}
		}
	} else {
		/* sending from cache */
		headbuf = xstrndup(connection->cachemap->data, connection->cachemap->offset);
	}

      parse_header:

	if (!strncasecmp(headbuf, "HTTP/", 5)) {
		/* pass the server's header through the rewrite rules */
		filebuf = filebuf_new();
		filebuf->data = headbuf;
		filebuf->size = strlen(headbuf) + 1;

		rewrite_do(rewrite_list, connection, filebuf, REWRITE_SERVER, TRUE);

		headbuf = filebuf->data;
		filebuf->data = NULL;

		filebuf_free(filebuf);
	}

	connection->rheader = http_header_parse_response(headbuf);

	if (connection->rheader == NULL) {
		if (postdata != NULL)
			filebuf_free(postdata);

		template_send(templates, "badresponse", connection, 400);
		xfree(headbuf);
		return -1;
	} else if (connection->rheader->version == HTTP_HTTP09) {
		/* HTTP/0.9 response with no header, push data back into socket buffer */
		sock_pushfront(connection->server, headbuf, strlen(headbuf));
	}

	xfree(headbuf);

	connection->rheader->host = xstrdup(connection->header->host);
	connection->rheader->file = xstrdup(connection->header->file);

	if (connection->cachemap != NULL) {
		if (connection->flags & CONNECTION_VALIDATE) {
			if (connection->rheader->last_modified != NULL && connection->rheader->code != 304) {
				/* if server doesn't understand the If-Modified-Since header,
				   compare the Last-Modified time with that of our cached copy */
				strptime(connection->rheader->last_modified, HTTPTIMEFORMAT, &tt);
				t = mktime(&tt);
			}

			if (connection->rheader->code == 304 || (connection->rheader->last_modified != NULL && t == connection->cachemap->lmtime)) {
				/* cached copy is up to date */
				if (connection->rheader->code != 304) {
					/* transfer in progress, we must close the connection */
					connection->keepalive_server = FALSE;
					sock_close(connection->server);
					connection->server = NULL;
				}

				connection->flags &= ~CONNECTION_VALIDATE;
				connection->flags |= CONNECTION_VALIDATED;
				cache_validated(global->cache, connection->cachemap);

				http_header_free(connection->rheader);

				headbuf = xstrndup(connection->cachemap->data, connection->cachemap->offset);

				putlog(MMLOG_CACHE, "validated: %s", connection->cachemap->key);

				goto parse_header;
			} else {
				/* expired */
				/* probably shouldn't count a cache that couldn't be verified as a hit */
				pthread_mutex_lock(&global->cache->lock);
				global->cache->hit--;
				global->cache->miss++;
				pthread_mutex_unlock(&global->cache->lock);

				putlog(MMLOG_CACHE, "failed validation: %s", connection->cachemap->key);

				cache_invalidate(global->cache, connection->cachemap);

				connection->cachemap = NULL;
			}
		}

		if (connection->cachemap != NULL) {
			snprintf(buf, sizeof(buf), "%u", (unsigned int) (utc_time(NULL) - connection->cachemap->mtime));
			FREE_AND_NULL(connection->rheader->age);
			connection->rheader->age = xstrdup(buf);

			if (!(connection->flags & CONNECTION_VALIDATED) && (atomic_read_mutex(&global->cache->lock, &connection->cachemap->flags) & CACHE_EXPIRED)) {
				/* sending an unvalidated stale cache file */
				FREE_AND_NULL(connection->rheader->warning);
				connection->rheader->warning = xstrdup("110 stale cache");
			}
		}

		if (transfer_limit_check(connection) == FALSE)
			goto out;
	} else {
		if (transfer_limit_check(connection) == FALSE)
			goto out;
	}

	if (connection->cachemap == NULL) {
		snprintf(buf, sizeof(buf), "%s://%s:%d%s", connection->header->proto, connection->header->host, connection->header->port, connection->header->file);
		if (cacheable(connection->header) && cacheable(connection->rheader))
			connection->cachemap = cache_open(connection, global->cache, buf, connection->rheader->raw, CACHE_WRITING | ((connection->flags & CONNECTION_PREFETCH) ? CACHE_PREFETCHED : 0));

		if (connection->cachemap != NULL) {
			putlog(MMLOG_CACHE, "create: %s", connection->cachemap->key);
			connection->flags |= CONNECTION_CACHING;
		} else if (connection->flags & CONNECTION_PREFETCH)
			goto out;

		xcache_header_update(connection, MISS);
	} else {

		/* the header is stored in the cache file as received from the server, so it needs to be
		   fixed up a bit */
		connection->rheader->content_length = connection->cachemap->size - connection->cachemap->offset;
		connection->rheader->chunked = FALSE;

		xcache_header_update(connection, HIT);
	}

	if (connection->rheader->code == 100) {
		/* server sent notice to continue POST, go back and retrieve final header 
		   (according to the RFC, the server should not send this unless the client sends an Expect:
		   header; however... IIS will sometimes send this anyways)
		 */
		http_header_free(connection->rheader);

		goto reget_header;
	}

	keepalive_check(connection);

	if (connection->rheader->code == 407 && connection->proxy_type == PROXY_NORMAL && !(connection->flags & CONNECTION_AUTHENTICATED)) {
		/* authenticate with the other proxy */
		if (proxy_authenticate(connection)) {
			http_header_free(connection->rheader);

			/* only try once */
			connection->flags |= CONNECTION_AUTHENTICATED;

			/* authentication finished, retrieve the header again */
			goto auth_finished;
		}
		FREE_AND_NULL(connection->rheader->proxy_authenticate);

		/* just continue, user will see 407 warning from proxy in browser */
	}

	if (postdata != NULL)
		filebuf_free(postdata);

	/* connection to other proxy when fowarding could have been lost throughout NTLM authentication */
	if (connection->server == NULL && connection->cachemap == NULL)
		goto incomplete;

	/* send template instead of web server's content when a template name matches the code */
	snprintf(buf, sizeof(buf), "%d", connection->rheader->code);
	ret = template_send(templates, buf, connection, connection->rheader->code);
	if (ret == TRUE)
		goto incomplete;

	if (connection->rheader->location != NULL && (connection->rheader->code == 301 || connection->rheader->code == 302)) {
		/* redirect_do will replace the Location: header if any rules match */
		redirect_do(redirect_list, connection, REDIRECT_HEADER);

		/* add URL command to location: header so the feature(s) are still bypassed when
		   browser follows redirect */
		if (connection->url_command != NULL) {
			ptr = url_command_add(connection, connection->rheader->location);
			xfree(connection->rheader->location);
			connection->rheader->location = ptr;
		}
	}

	connection->rheader->header_filtered = header_filter(header_list, connection, HEADER_SERVER);

	/* no message body regardless of what header says for HEAD requests and certain return codes */
	if (!strcasecmp(connection->header->method, "HEAD") || connection->rheader->code == 304 || connection->rheader->code == 204 || (connection->rheader->code >= 100 && connection->rheader->code < 200)) {
		header_send(connection->rheader, connection, CLIENT, HEADER_RESP);

		goto incomplete;
	}

	if (url_command_find(connection->url_command, "mime")) {
		/* show matching MIME filter entries, if any, for the requested URL */
		mime_check_show(connection);

		goto incomplete;
	}

	if (url_command_find(connection->url_command, "headers")) {
		show_headers(connection);

		goto incomplete;
	}

	mime_list_list = mime_check(mime_list, connection);
	if (mime_list_list != NULL) {
		putlog(MMLOG_MIME, "blocked mime-type %s from %s%s", (connection->rheader->content_type != NULL) ? connection->rheader->content_type : "(none)", connection->header->host, connection->header->file);

		template_send(templates, (mime_list_list->template != NULL) ? mime_list_list->template : (mime_list->dtemplate != NULL) ? mime_list->dtemplate : "blocked", connection, 200);

		pthread_rwlock_unlock(&mime_list->lock);

		goto incomplete;
	}
	pthread_rwlock_unlock(&mime_list->lock);

	putlog((connection->flags & CONNECTION_PREFETCH) ? MMLOG_PREFETCH : MMLOG_REQUEST, "%s %s://%s:%d%s", connection->header->method, connection->header->proto, connection->header->host, connection->header->port, connection->header->file);

	prefetch_setup_callbacks(connection, global->prefetch);

	if (connection->flags & (CONNECTION_PROCESS | CONNECTION_PREFETCH)) {
		if (buffer_check(connection)) {
			connection->flags |= CONNECTION_BUFFER;
			filebuf_size = atomic_read_rwlock(&global->general->lock, &global->general->maxbuffer);
		}
	}

	putlog(MMLOG_DEBUG, "buffering: %d processing: %d", !!(connection->flags & CONNECTION_BUFFER), !!(connection->flags & CONNECTION_PROCESS));

	if (!(connection->flags & CONNECTION_BUFFER)) {
		header_send(connection->rheader, connection, CLIENT, HEADER_RESP);

		if (connection->cachemap != NULL && !(connection->flags & CONNECTION_CACHING)) {
			if (connection->htmlstream != NULL) {
				htmlstream_add(connection->htmlstream, connection->cachemap->data, connection->cachemap->size);
				htmlstream_add(connection->htmlstream, NULL, 0);
			}

			net_cache_send(connection->cachemap, connection, CLIENT, 0);
		} else
			http_transfer(connection, SERVER);
	} else {
		/* this file is to be buffered and processed before sending it to the browser */

		if (!(connection->flags & CONNECTION_CACHING) && connection->cachemap != NULL)
			filebuf = cache_to_filebuf(connection->cachemap, filebuf_size);
		else
			filebuf = http_transfer_filebuf(connection, SERVER, filebuf_size);

		filebuf_size = filebuf->size;
		content_length = connection->rheader->content_length;
		chunked = connection->rheader->chunked;

		/* no point doing anything if page is empty */
		if (filebuf_size == 0)
			goto bypass;

		if (connection->flags & CONNECTION_CACHING)
			cache_add(global->cache, connection->cachemap, filebuf->data, filebuf_size);

		/* decompress gzip-encoded content */
		if (connection->rheader->content_encoding != NULL) {
			if (content_length != filebuf_size) {
				/* file is bigger than buffer, just send as-is */
				putlog(MMLOG_WARN, "too big file width content encoding: %s", connection->rheader->content_encoding);

				goto bypass;
			}

			if (filebuf_decode(connection, filebuf) == FALSE)
				goto bypass;

			FREE_AND_NULL(connection->rheader->content_encoding);
		}

		/* check keyword score */
		if (url_command_find(connection->url_command, "score")) {
			ret = keyword_check(keyword_list, connection, filebuf, TRUE);

			score_show(connection, ret);

			filebuf_free(filebuf);

			goto out;
		}

		if (url_command_find(connection->url_command, "diff")) {
			filebuf2 = filebuf_dup(filebuf);

			rewrite_do(rewrite_list, connection, filebuf2, REWRITE_BODY, TRUE);

			show_filebuf_diff(connection, filebuf, filebuf2);

			filebuf_free(filebuf);
			filebuf_free(filebuf2);

			goto out;
		}

		if (connection->flags & CONNECTION_PROCESS) {
			ret = keyword_check(keyword_list, connection, filebuf, TRUE);

			if (ret) {
				pthread_rwlock_rdlock(&keyword_list->lock);
				if (ret >= keyword_list->threshold) {
					/* page blocked due to keyword score exceeding threshold */
					putlog(MMLOG_KEYWORDS, "page above threshold with a score of %d", ret);

					template_send(templates, (keyword_list->template != NULL) ? keyword_list->template : "blocked", connection, 200);

					pthread_rwlock_unlock(&keyword_list->lock);

					filebuf_free(filebuf);

					goto out;
				}

				pthread_rwlock_unlock(&keyword_list->lock);
			}

			rewrite_do(rewrite_list, connection, filebuf, REWRITE_BODY, TRUE);

			external_list = external_find(external, connection);
			if (external_list != NULL && external_list->exec != NULL) {
				filebuf2 = external_exec(connection, external_list->exec, filebuf, external_list->type);
				if (filebuf2 != NULL) {
					filebuf_free(filebuf);
					filebuf = filebuf2;

					if (external_list->newmime != NULL) {
						FREE_AND_NULL(connection->rheader->content_type);

						if (!strcasecmp(external_list->newmime, "STDIN")) {
							/* try extracting Content-type header from external 
							   program's output */
							connection->rheader->content_type = external_getmime(filebuf);
						} else
							connection->rheader->content_type = xstrdup(external_list->newmime);
					}
				} else
					putlog(MMLOG_ERROR, "failed to execute external parser '%s'", external_list->exec);
			}
			pthread_rwlock_unlock(&external->lock);

		}

		if (url_command_find(connection->url_command, "raw")) {
			show_filebuf(connection, filebuf);

			filebuf_free(filebuf);

			goto out;
		}

		if (connection->htmlstream != NULL) {
			htmlstream_add(connection->htmlstream, filebuf->data, filebuf->size);
			htmlstream_add(connection->htmlstream, NULL, 0);
		}

		if (url_command_find(connection->url_command, "htmltree")) {
			if (connection->htmlstream == NULL) {
				connection->htmlstream = htmlstream_new();
				htmlstream_add(connection->htmlstream, filebuf->data, filebuf->size);
				htmlstream_add(connection->htmlstream, NULL, 0);
			}

			show_htmltree(connection);

			filebuf_free(filebuf);

			goto out;
		}

		filebuf_encode(connection, filebuf);

	      bypass:

		if (content_length == filebuf_size) {
			if (connection->keepalive_client != FALSE && connection->header->keepalive != FALSE && connection->header->proxy_keepalive != FALSE) {
				/* we may be able to do keep-alive now since the page was buffered and the size is now known */
				if (connection->header->keepalive == TRUE || connection->header->proxy_keepalive == TRUE)
					connection->keepalive_client = TRUE;
				else if (connection->header->version == HTTP_HTTP11)
					connection->keepalive_client = TRUE;
			}
		} else if (content_length == -1)
			connection->keepalive_client = FALSE;

		connection->rheader->chunked = FALSE;

		if (content_length != -1)
			connection->rheader->content_length += filebuf->size - filebuf_size;

		putlog(MMLOG_DEBUG, "orig. filebuf size: %d, orig. content-length: %d", filebuf_size, content_length);
		putlog(MMLOG_DEBUG, "new. filebuf size: %d, new. content-length: %d", filebuf->size, connection->rheader->content_length);

		header_send(connection->rheader, connection, CLIENT, HEADER_RESP);
		net_filebuf_send(filebuf, connection, CLIENT);

		filebuf_free(filebuf);

		if (content_length != filebuf_size) {
			/* transfer rest of response */
			if (connection->cachemap != NULL && !(connection->flags & CONNECTION_CACHING))
				net_cache_send(connection->cachemap, connection, CLIENT, filebuf_size);
			else {
				if (content_length != -1)
					net_transfer(connection, SERVER, content_length - filebuf_size);
				else if (chunked)
					do {
						ret = next_chunksize(connection, SERVER);
						if (net_transfer(connection, SERVER, ret) != ret)
							break;
						/* don't allow extra \r\n in cache file */
						sock_getline(connection->server, NULL, -1, atomic_read_rwlock(&global->general->lock, &global->general->timeout));
					} while (ret > 0);
				else
					net_transfer(connection, SERVER, -1);
			}
		}
	}

	http_header_free(connection->rheader);

	return TRUE;

      incomplete:

	if (connection->cachemap != NULL) {
		cache_invalidate(global->cache, connection->cachemap);
		connection->cachemap = NULL;
	}

      out:

	connection->keepalive_server = FALSE;
	http_header_free(connection->rheader);

	return TRUE;
}

/*
This function handles CONNECT requests (HTTPS)
*/
int protocol_connect(CONNECTION * connection)
{
	char *headbuf;

	putlog(MMLOG_REQUEST, "CONNECT %s:%d", connection->header->host, connection->header->port);

	if (connection->proxy_type == PROXY_NORMAL) {
		header_send(connection->header, connection, SERVER, HEADER_FORWARD);

	      auth_finished:
		headbuf = header_get(connection, SERVER, atomic_read_rwlock(&global->general->lock, &global->general->timeout));

		if (headbuf != NULL) {
			connection->rheader = http_header_parse_response(headbuf);
			xfree(headbuf);
		}

		if (connection->rheader == NULL) {
			template_send(templates, "noconnect", connection, 503);

			goto out;
		}

		if (connection->rheader->code == 407 && !(connection->flags & CONNECTION_AUTHENTICATED)) {
			if (proxy_authenticate(connection)) {
				http_header_free(connection->rheader);

				/* only try once */
				connection->flags |= CONNECTION_AUTHENTICATED;

				/* authentication finished, retrieve the header again */
				goto auth_finished;
			}

			FREE_AND_NULL(connection->rheader->proxy_authenticate);
		} else if (connection->rheader->code != 200) {
			template_send(templates, "noconnect", connection, 503);

			goto out;
		}

	}

	putsock(connection->client, "HTTP/1.1 200 Connection Established\r\n\r\n");

	net_proxy(connection, -1);

      out:
	if (connection->rheader != NULL)
		http_header_free(connection->rheader);
	return 0;
}

/*
get size of next chunk for pages using chunked encoding
*/
unsigned int next_chunksize(CONNECTION * connection, int flags)
{
	int x;
	char buf[64];

	x = sock_getline((flags == SERVER) ? connection->server : connection->client, buf, sizeof(buf), atomic_read_rwlock(&global->general->lock, &global->general->timeout));
	if (x <= 0)
		return 0;

	return strtol(buf, NULL, 16);
}


/*
transfer http body from one socket to another
*/
void http_transfer(CONNECTION * connection, int direction)
{
	int x;
	HEADER *header;
	SOCKET *sock, *sock2;

	header = (direction == SERVER) ? connection->rheader : connection->header;
	sock = (direction == SERVER) ? connection->server : connection->client;
	sock2 = (direction == SERVER) ? connection->client : connection->server;

	if (header->content_length != -1)
		net_transfer(connection, direction, header->content_length);
	else if (header->chunked)
		do {
			x = next_chunksize(connection, direction);
			if (sock2 != NULL)
				putsock(sock2, "%x\r\n", x);
			if (net_transfer(connection, direction, x) != x)
				break;
			/* don't allow extra \r\n in cache file */
			sock_getline(sock, NULL, -1, atomic_read_rwlock(&global->general->lock, &global->general->timeout));
			if (sock2 != NULL)
				putsock(sock2, "\r\n");
		} while (x > 0);
	else
		net_transfer(connection, direction, -1);

}

/*
transfer http body into filebuf
*/
FILEBUF *http_transfer_filebuf(CONNECTION * connection, int direction, int maxlen)
{
	int x, size;
	FILEBUF *filebuf;
	HEADER *header;
	SOCKET *sock;

	filebuf = filebuf_new();

	header = (direction == SERVER) ? connection->rheader : connection->header;
	sock = (direction == SERVER) ? connection->server : connection->client;

	if (header->content_length != -1) {
		size = (maxlen == -1 || maxlen > header->content_length) ? header->content_length : maxlen;
		net_filebuf_read(filebuf, connection, direction, size);

	} else if (header->chunked) {
		size = 0;
		do {
			x = next_chunksize(connection, direction);
			if (!net_filebuf_read(filebuf, connection, direction, x))
				break;
			/* discard the \r\n from the server */
			sock_getline(sock, NULL, -1, atomic_read_rwlock(&global->general->lock, &global->general->timeout));

			size += x;
		} while (x > 0 && size < maxlen);

		if (x == 0)
			header->content_length = filebuf->size;
	} else {
		x = net_filebuf_read(filebuf, connection, direction, maxlen);
		if (x == FALSE)
			header->content_length = filebuf->size;
	}

	return filebuf;
}

/* 
read the body and discard it 
*/
void http_transfer_discard(CONNECTION * connection, int direction)
{
	int x;
	HEADER *header;
	SOCKET *sock;

	header = (direction == SERVER) ? connection->rheader : connection->header;
	sock = (direction == SERVER) ? connection->server : connection->client;

	if (header->content_length != -1)
		net_filebuf_read(NULL, connection, direction, header->content_length);
	else if (header->chunked)
		do {
			x = next_chunksize(connection, direction);
			net_filebuf_read(NULL, connection, direction, x);
			/* discard the \r\n from the server */
			sock_getline(sock, NULL, -1, atomic_read_rwlock(&global->general->lock, &global->general->timeout));
		} while (x > 0);
	else
		net_filebuf_read(NULL, connection, direction, -1);

}

int proxy_authenticate(CONNECTION * connection)
{
	if (connection->rheader->proxy_authenticate == NULL)
		return FALSE;

	if (!strcasecmp(connection->rheader->proxy_authenticate, "NTLM")) {
		/* NTLM authentication */
#ifdef ENABLE_NTLM
		return send_ntlm_response(connection);
#endif
	} else if (!strncasecmp(connection->rheader->proxy_authenticate, "Basic", 5)) {
		/* Basic authentication */
		return send_basic_response(connection);
	}

	return FALSE;
}

#ifdef ENABLE_NTLM
/* perform an NTLM handshake, the sequence of events looks like this:
    1: C -->  S   GET ...

    2: C <--  S   401 Unauthorized
                  Proxy-Authenticate: NTLM
    
    3: C  --> S   GET ...
                  Proxy-Authorization: NTLM <base64-encoded type-1-message>
    
    4: C <--  S   401 Unauthorized
                  Proxy-Authenticate: NTLM <base64-encoded type-2-message>
    
    5: C  --> S   GET ...
                  Proxy-Authorization: NTLM <base64-encoded type-3-message>
    
    6: C <--  S   200 Ok
*/
int send_ntlm_response(CONNECTION * connection)
{
	int ret, oldkeep, oldcontentlength;
	unsigned char buf[4096], buf2[4096], *headbuf, *oldmethod;
	static pthread_mutex_t ntlmlock = PTHREAD_MUTEX_INITIALIZER;
	tSmbNtlmAuthRequest request;
	tSmbNtlmAuthChallenge challenge;
	tSmbNtlmAuthResponse response;
	HEADER *header;

	/* client must do keep-alive during authentication */
	oldkeep = connection->keepalive_server;
	connection->keepalive_server = TRUE;

	/* must use GET method for handshake */
	oldmethod = connection->header->method;
	connection->header->method = xstrdup("GET");
	oldcontentlength = connection->header->content_length;
	connection->header->content_length = -1;

	/* don't need the body of the 407 message */
	http_transfer_discard(connection, SERVER);

	/* send type-1 message */
	/* note: buildSmbNtml* knows how to deal with NULL arguments */
	pthread_mutex_lock(&ntlmlock);
	buildSmbNtlmAuthRequest(&request, connection->proxy_username, connection->proxy_domain);
	pthread_mutex_unlock(&ntlmlock);

	to64frombits(buf, (char *) &request, SmbLength(&request));

	FREE_AND_NULL(connection->proxy_auth);
	snprintf(buf2, sizeof(buf2), "NTLM %s", buf);
	connection->proxy_auth = xstrdup(buf2);

	header_send(connection->header, connection, SERVER, HEADER_FORWARD);

	headbuf = header_get(connection, SERVER, atomic_read_rwlock(&global->general->lock, &global->general->timeout));
	if (headbuf == NULL) {
		/* squid and maybe some other proxies will disconnect after sending
		   the first 407 message regardless of Connection header */
		do {
			ret = protocol_reconnect(connection);
			if (ret == -1)
				goto error;

			header_send(connection->header, connection, SERVER, HEADER_FORWARD);

			headbuf = header_get(connection, SERVER, atomic_read_rwlock(&global->general->lock, &global->general->timeout));
		} while (headbuf == NULL && connection->keepalive_server == TRUE);

		if (headbuf == NULL)
			goto error;
	}

	header = http_header_parse_response(headbuf);
	xfree(headbuf);
	if (header == NULL)
		goto error;

	http_header_free(connection->rheader);
	connection->rheader = header;

	if (header->proxy_authenticate == NULL || strncasecmp(header->proxy_authenticate, "NTLM ", 5))
		goto error;

	http_transfer_discard(connection, SERVER);

	/* parse type-2 message */
	from64tobits((char *) &challenge, &header->proxy_authenticate[5]);

	pthread_mutex_lock(&ntlmlock);
	buildSmbNtlmAuthResponse(&challenge, &response, connection->proxy_username, connection->proxy_password);
	pthread_mutex_unlock(&ntlmlock);

	/* send type-3 message */
	to64frombits(buf, (char *) &response, SmbLength(&response));
	snprintf(buf2, sizeof(buf2), "NTLM %s", buf);

	FREE_AND_NULL(connection->proxy_auth);
	connection->proxy_auth = xstrdup(buf2);

	connection->keepalive_server = oldkeep;
	xfree(connection->header->method);
	connection->header->method = oldmethod;
	connection->header->content_length = oldcontentlength;

	header_send(connection->header, connection, SERVER, HEADER_FORWARD);

	FREE_AND_NULL(connection->proxy_auth);

	return TRUE;

      error:
	connection->keepalive_server = oldkeep;
	xfree(connection->header->method);
	connection->header->method = oldmethod;
	connection->header->content_length = oldcontentlength;

	return FALSE;
}

#endif /* ENABLE_NTLM */

/* basic proxy authentication */
int send_basic_response(CONNECTION * connection)
{
	char buf[4096], buf2[4096];

	/* discard body of 407 authentication required message */
	http_transfer_discard(connection, SERVER);

	snprintf(buf, sizeof(buf), "%s:%s", (connection->proxy_username != NULL) ? connection->proxy_username : "", (connection->proxy_password != NULL) ? connection->proxy_password : "");
	to64frombits(buf2, buf, strlen(buf));

	snprintf(buf, sizeof(buf), "Basic %s", buf2);

	FREE_AND_NULL(connection->proxy_auth);
	connection->proxy_auth = xstrdup(buf);

	header_send(connection->header, connection, SERVER, HEADER_FORWARD);

	FREE_AND_NULL(connection->proxy_auth);

	return TRUE;
}

int protocol_ftp(CONNECTION * connection)
{
	int x, lastline = -1, first = TRUE, code, timeout, order = -1, field = -1;
	char *ptr, *ptr2, buf[1024], *action = NULL, *argument = NULL;
	FILEBUF *filebuf = NULL;
	struct dir_t *dir, *dstart;
	struct cgi_args_t *cgiargs = NULL, *cgitmp;
	SOCKET *sock;

	timeout = atomic_read_rwlock(&global->ftp->lock, &global->ftp->timeout);

	putlog((connection->flags & CONNECTION_PREFETCH) ? MMLOG_PREFETCH : MMLOG_REQUEST, "%s ftp://%s:%d%s", connection->header->method, connection->header->host, connection->header->port, connection->header->file);

	/* look for cgi arguments */
	ptr = strchr(connection->header->file, '?');
	if (ptr != NULL) {
		cgiargs = cgi_parse_request(&ptr[1]);
		*ptr = '\0';
	}

	if (connection->header->username == NULL) {
		pthread_rwlock_rdlock(&global->ftp->lock);
		connection->header->username = xstrdup(global->ftp->anonlogin);
		connection->header->password = xstrdup(global->ftp->anonpass);
		pthread_rwlock_unlock(&global->ftp->lock);
	}

	if (!(connection->flags & CONNECTION_LOGGEDIN)) {
	      login:
		putsock(connection->server, "USER %s\r\n", connection->header->username);
		x = ftp_code_wait(connection, 331, 530);
		if (x == 530) {
			template_send(templates, "badauth", connection, 404);
			goto error;
		} else if (x == -1) {
			if (connection->keepalive_server == TRUE) {
				x = protocol_reconnect(connection);
				if (x == -1)
					goto error;

				if (!(connection->flags & CONNECTION_LOGGEDIN))
					/* fresh connection, not from pool */
					goto login;
			} else {
				template_send(templates, "noconnect", connection, 404);
				goto error;
			}
		}

		if (!(connection->flags & CONNECTION_LOGGEDIN)) {
			putsock(connection->server, "PASS %s\r\n", connection->header->password);
			x = sock_getline(connection->server, buf, sizeof(buf), timeout);
			if (x <= 0 || atoi(buf) == 530) {
				template_send(templates, "badauth", connection, 404);
				goto error;
			}
		}
	}

	putsock(connection->server, "TYPE I\r\n");
	x = ftp_code_wait(connection, 200, 0);
	if (x == -1) {
		if (connection->flags & CONNECTION_LOGGEDIN) {
			/* this connection is from the pool, keep on reconnecting until 
			   we get a working pool connection or a fresh connection */
			connection->flags &= ~CONNECTION_LOGGEDIN;
			goto login;
		} else {
			template_send(templates, "noconnect", connection, 404);
			goto error;
		}
	}

	connection->keepalive_server = TRUE;
	connection->flags |= CONNECTION_LOGGEDIN;

	if (connection->header->content_length != -1) {
		/* client wants to upload something */

		/* for FILE uploads, the body of the requeset will look like this:
		   ----------SOMERANDOMGIBBERISH
		   Content-Disposition: form-data; name="upload"; filename="somefile"
		   Content-Type: application/octet-stream
		   \r\n
		   <FILE CONTENTS>
		   \r\n
		   ----------SAMEGIBBERISHABOVE--
		 */

		/* \r\n after file contents and extra -- after gibberish */
		connection->header->content_length -= 4;

		while ((x = sock_getline(connection->client, buf, sizeof(buf), -1)) != -1) {
			connection->header->content_length -= strlen(buf) * ((first == TRUE) ? 2 : 1);
			if (first == TRUE)
				lastline = strlen(buf);
			first = FALSE;

			if (!strncasecmp(buf, "Content-Disposition:", 20)) {
				ptr = s_strcasestr(buf, "filename");
				if (ptr != NULL) {
					ptr = strchr(ptr, '\"');
					if (ptr != NULL) {
						ptr2 = strchr(&ptr[1], '\"');
						if (ptr2 != NULL)
							*ptr2 = '\0';
						ptr++;

						if (connection->header->file[strlen(connection->header->file) - 1] != '/')
							connection->header->file = string_append(connection->header->file, "/");

						connection->header->file = string_append(connection->header->file, ptr);

						putlog(MMLOG_DEBUG, "ftp upload filename: %s", connection->header->file);
					}
				}
			} else if ((*buf == '\r' && buf[1] == '\n') || *buf == '\n')
				break;

		}
		putlog(MMLOG_DEBUG, "ftp upload file length: %d", connection->header->content_length);

		if (x == -1 || connection->header->content_length < 0) {
			connection->keepalive_client = FALSE;

			goto error;
		}

		sock = ftp_transfer_setup(connection, "STOR %s\r\n", connection->header->file);
		if (sock == NULL) {
			/* discard request body */
			http_transfer_discard(connection, CLIENT);

			template_send(templates, "noaccess", connection, 403);

			goto error;
		}

		x = ftp_transfer(connection, CLIENT, sock, connection->header->content_length);
		if (x == -1)
			connection->keepalive_client = FALSE;
		else {
			if (lastline != -1) {
				/* client uploaded a file, read in the last two lines after the
				   file body */
				sock_getline(connection->client, NULL, -1, -1);
				sock_getline(connection->client, NULL, -1, -1);
			}
		}

		sock_close(sock);

		/* grab listing of directory we just uploaded file to */
		ptr = strrchr(connection->header->file, '/');
		if (ptr != NULL)
			*ptr = '\0';
	}

	putsock(connection->server, "CWD %s\r\n", connection->header->file);
	code = ftp_code_wait(connection, 250, 550);
	if (code == -1) {
		template_send(templates, "noconnect", connection, 404);
		goto error;
	} else if (code == 250) {
		/* ok... user wants directory listing */

		if (connection->cachemap != NULL) {
			/* I guess this could happen if the proxy was previously
			   setup to forward FTP requests through another proxy,
			   and the index was cached.. may as well get rid of it */
			cache_invalidate(global->cache, connection->cachemap);
			connection->cachemap = NULL;
		}

		if (connection->header->file[strlen(connection->header->file) - 1] != '/') {
			/* request for directory listing without trailing '/' on URL, 
			   we need to send a redirect so relative links work properly */

			connection->header->file = string_append(connection->header->file, "/");

			connection->rheader = header_new();
			connection->rheader->type = HTTP_RESP;
			connection->rheader->code = 302;
			connection->rheader->content_length = 0;
			connection->rheader->location = xstrdup(connection->header->file);

			header_send(connection->rheader, connection, CLIENT, HEADER_RESP);

			goto out;
		}

		if (cgiargs != NULL) {
			for (cgitmp = cgiargs; cgitmp; cgitmp = cgitmp->next) {
				if (!strcasecmp(cgitmp->name, "action"))
					action = cgitmp->value;
				else if (!strcasecmp(cgitmp->name, "argument"))
					argument = cgitmp->value;
			}

			if (action != NULL) {
				if (argument != NULL) {
					if (!strcasecmp(action, "mkdir"))
						putsock(connection->server, "MKD %s\r\n", argument);
					else if (!strcasecmp(action, "raw"))
						putsock(connection->server, "%s\r\n", argument);
				}

				for (cgitmp = cgiargs; cgitmp; cgitmp = cgitmp->next) {
					if (!strcasecmp(cgitmp->name, "filename")) {
						if (!strcasecmp(action, "delete"))
							putsock(connection->server, "DELE %s\r\n", cgitmp->value);
						else if (!strcasecmp(action, "rmdir"))
							putsock(connection->server, "RMD %s\r\n", cgitmp->value);

						if (argument != NULL) {
							if (!strcasecmp(action, "rename")) {
								putsock(connection->server, "RNFR %s\r\n", cgitmp->value);
								putsock(connection->server, "RNTO %s\r\n", argument);
							}
						}
					}
				}
			}
		}


		sock = ftp_transfer_setup(connection, "LIST\r\n");
		if (sock != NULL)
			filebuf = ftp_transfer_filebuf(connection, SERVER, sock, -1);
		if (sock == NULL || filebuf == NULL) {
			if (sock != NULL)
				sock_close(sock);
			template_send(templates, "noconnect", connection, 404);
			goto error;
		}
		sock_close(sock);

		if (url_command_find(connection->url_command, "raw")) {
			show_filebuf(connection, filebuf);

			filebuf_free(filebuf);

			goto out;
		}


		dir = dir_list_parse(filebuf);

		filebuf_free(filebuf);

		if (action != NULL && argument != NULL && !strcasecmp(action, "rematch"))
			dir = dir_list_filter(dir, argument);

		if (cgiargs != NULL) {
			for (cgitmp = cgiargs; cgitmp; cgitmp = cgitmp->next) {
				if (!strcasecmp(cgitmp->name, "order")) {
					if (!strcasecmp(cgitmp->value, "ascending"))
						order = SORT_ASCENDING;
					else if (!strcasecmp(cgitmp->value, "descending"))
						order = SORT_DESCENDING;
					else
						order = -1;
				} else if (!strcasecmp(cgitmp->name, "field")) {
					if (!strcasecmp(cgitmp->value, "name"))
						field = SORT_NAME;
					else if (!strcasecmp(cgitmp->value, "size"))
						field = SORT_SIZE;
					else if (!strcasecmp(cgitmp->value, "date"))
						field = SORT_DATE;
					else
						field = -1;
				}
			}
		}

		if (order != -1 && field != -1)
			dir = dir_list_sort(dir, order, field);
		else {
			order = atomic_read_rwlock(&global->ftp->lock, &global->ftp->sortorder);
			field = atomic_read_rwlock(&global->ftp->lock, &global->ftp->sortfield);

			if (field != SORT_NONE)
				dir = dir_list_sort(dir, order, field);
		}

		show_dirlist(connection, dir);
		dir_list_free(dir);
	} else if (code == 550) {
		ptr = strrchr(connection->header->file, '/');

		s_strncpy(buf, connection->header->file, (ptr - connection->header->file) + 2);
		putsock(connection->server, "CWD %s\r\n", buf);

		x = sock_getline(connection->server, buf, sizeof(buf), timeout);
		if (x <= 0)
			goto error;

		code = atoi(buf);
		if (code == 550) {
			template_send(templates, "nofile", connection, 404);
			goto error;
		} else if (code == 250) {
			sock = ftp_transfer_setup(connection, "LIST\r\n");
			if (sock != NULL)
				filebuf = ftp_transfer_filebuf(connection, SERVER, sock, -1);
			if (sock == NULL || filebuf == NULL) {
				if (sock != NULL)
					sock_close(sock);
				template_send(templates, "noconnect", connection, 404);
				goto error;
			}
			sock_close(sock);

			connection->rheader = header_new();
			connection->rheader->type = HTTP_RESP;
			connection->rheader->code = 200;

			dstart = dir = dir_list_parse(filebuf);

			for (; dir && strcmp(dir->name, &ptr[1]); dir = dir->next);
			if (dir == NULL) {
				/* file was not in directory listing */
				connection->rheader->content_length = -1;
				/* content length is not known, cannot do keep-alive */
				connection->keepalive_client = FALSE;

				if (connection->cachemap != NULL) {
					/* this may be a temporary error, keep the cache file
					   in case we can validate it later */
					cache_close(global->cache, connection->cachemap);
					connection->cachemap = NULL;
				}
			} else {
				strftime(buf, sizeof(buf), HTTPTIMEFORMAT, &dir->time);
				connection->rheader->last_modified = xstrdup(buf);
				connection->rheader->content_length = dir->size;
			}

			if (connection->cachemap != NULL) {
				/* must be validating it ... */
				if (mktime(&dir->time) == connection->cachemap->lmtime) {
					putlog(MMLOG_CACHE, "validated: %s", connection->cachemap->key);

					connection->flags &= ~CONNECTION_VALIDATE;
					connection->flags |= CONNECTION_VALIDATED;
					cache_validated(global->cache, connection->cachemap);

					http_header_free(connection->rheader);
					connection->rheader = http_header_dup(connection->cachemap->header);

					snprintf(buf, sizeof(buf), "%u", (unsigned int) (utc_time(NULL) - connection->cachemap->mtime));
					FREE_AND_NULL(connection->rheader->age);
					connection->rheader->age = xstrdup(buf);

					dir_list_free(dstart);

					if (transfer_limit_check(connection) == FALSE)
						goto out;

					xcache_header_update(connection, HIT);

					header_send(connection->rheader, connection, CLIENT, HEADER_RESP);
					net_cache_send(connection->cachemap, connection, CLIENT, 0);

					goto out;
				} else {
					/* don't count it as a hit if it failed validation */
					pthread_mutex_lock(&global->cache->lock);
					global->cache->hit--;
					global->cache->miss++;
					pthread_mutex_unlock(&global->cache->lock);

					putlog(MMLOG_CACHE, "validation failed: %s", connection->cachemap->key);

					cache_invalidate(global->cache, connection->cachemap);
					connection->cachemap = NULL;
				}
			}

			/* if we can't get a directory listing, we don't have enough 
			   information about the file to cache it. */
			if (dir != NULL && !strcasecmp(connection->header->username, ANONLOGIN)) {
				snprintf(buf, sizeof(buf), "%s://%s:%d%s", connection->header->proto, connection->header->host, connection->header->port, connection->header->file);

				filebuf = filebuf_new();
				header_send_filebuf(filebuf, connection->rheader, connection, CLIENT, HEADER_RESP);
				filebuf_add(filebuf, "", 1);
				connection->cachemap = cache_open(connection, global->cache, buf, filebuf->data, CACHE_WRITING | ((connection->flags & CONNECTION_PREFETCH) ? CACHE_PREFETCHED : 0));
				filebuf_free(filebuf);

				if (connection->cachemap != NULL)
					connection->flags |= CONNECTION_CACHING;
			}

			dir_list_free(dstart);

			if (transfer_limit_check(connection) == FALSE)
				goto out;

			sock = ftp_transfer_setup(connection, "RETR %s\r\n", connection->header->file);
			if (sock == NULL) {
				if (connection->cachemap != NULL) {
					cache_invalidate(global->cache, connection->cachemap);
					connection->cachemap = NULL;
				}

				template_send(templates, "nofile", connection, 404);

				goto error;
			}

			xcache_header_update(connection, MISS);

			header_send(connection->rheader, connection, CLIENT, HEADER_RESP);

			x = ftp_transfer(connection, SERVER, sock, connection->rheader->content_length);

			if (x == -1)
				connection->keepalive_client = FALSE;

			sock_close(sock);
		} else {
			template_send(templates, "badresponse", connection, 404);
			goto error;
		}
	}

      out:

	if (connection->rheader != NULL)
		http_header_free(connection->rheader);

	if (cgiargs != NULL)
		cgi_args_free(cgiargs);

	return TRUE;

      error:

	if (connection->rheader != NULL)
		http_header_free(connection->rheader);

	connection->keepalive_server = FALSE;

	if (cgiargs != NULL)
		cgi_args_free(cgiargs);

	return -1;
}

int ftp_code_wait(CONNECTION * connection, int scode, int fcode)
{
	int x, code = -1, timeout;
	char buf[128];

	timeout = atomic_read_rwlock(&global->ftp->lock, &global->ftp->timeout);

	while ((x = sock_getline(connection->server, buf, sizeof(buf), timeout)) > 0) {
		code = atoi(buf);

		if (code == scode || code == fcode)
			return code;
	}

	return code;
}

SOCKET *ftp_transfer_setup(CONNECTION * connection, char *fmt, ...)
{
	int passive, code, x, fd = -1, timeout;
	char buf[1024], cmd[256], *ptr;
	socklen_t socklen = sizeof(struct sockaddr_in);
	struct sockaddr_in *saddr = NULL;
	struct pollfd pfd;
	va_list valist;

	va_start(valist, fmt);
	vsnprintf(cmd, sizeof(cmd), fmt, valist);
	va_end(valist);

	passive = atomic_read_rwlock(&global->ftp->lock, &global->ftp->pasv);
	timeout = atomic_read_rwlock(&global->ftp->lock, &global->ftp->timeout);

	if (passive == TRUE) {
		putsock(connection->server, "PASV\r\n");

		while ((x = sock_getline(connection->server, buf, sizeof(buf), timeout)) > 0) {
			code = atoi(buf);
			if (code == 227)
				break;
		}
		if (x == -1)
			goto error;

		ptr = strchr(buf, '(');
		if (ptr == NULL)
			goto error;

		saddr = ftpport_parse(&ptr[1]);

		fd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
		if (fd == -1)
			goto error;
	} else {
		saddr = xmalloc(sizeof(struct sockaddr_in));
		memset(saddr, 0, sizeof(struct sockaddr_in));

		/* bind to same interface connection to ftp server is made on */
		getsockname(connection->server->fd, (struct sockaddr *) saddr, &socklen);
		saddr->sin_port = 0;

		fd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
		if (fd == -1)
			goto error;

		x = bind(fd, (struct sockaddr *) saddr, sizeof(struct sockaddr_in));
		if (x == -1)
			goto error;

		x = listen(fd, 1);
		if (x == -1)
			goto error;

		/* get the port the kernel chose for us */
		getsockname(fd, (struct sockaddr *) saddr, &socklen);

		ptr = ftpport_create(saddr);
		putsock(connection->server, "PORT %s\r\n", ptr);
		xfree(ptr);

		x = ftp_code_wait(connection, 200, 0);
		if (x == -1)
			goto error;
	}

	/* ok, tranfer is ready to go... */
	putsock(connection->server, "%s", cmd);

	if (passive == TRUE) {
		x = connect(fd, (struct sockaddr *) saddr, sizeof(struct sockaddr_in));
		if (x == -1)
			goto error;
	} else {
		pfd.fd = fd;
		pfd.events = POLLIN;

		x = p_poll(&pfd, 1, atomic_read_rwlock(&global->ftp->lock, &global->ftp->timeout) * 1000);
		if (x <= 0 || !(pfd.revents & POLLIN))
			goto error;

		x = accept(fd, (struct sockaddr *) saddr, &socklen);
		if (x == -1)
			goto error;

		close(fd);
		fd = x;
	}

	xfree(saddr);

	return sock_new(fd);

      error:

	if (saddr != NULL)
		xfree(saddr);
	if (fd != -1)
		close(fd);

	return NULL;
}

int ftp_transfer(CONNECTION * connection, int which, SOCKET * sock, int bytes)
{
	int x;
	SOCKET *sock2;

	sock2 = connection->server;

	connection->server = sock;

	x = net_transfer(connection, which, bytes);

	connection->server = sock2;

	return x;
}

FILEBUF *ftp_transfer_filebuf(CONNECTION * connection, int which, SOCKET * sock, int bytes)
{
	SOCKET *sock2;
	FILEBUF *filebuf;

	sock2 = connection->server;

	connection->server = sock;

	filebuf = filebuf_new();

	net_filebuf_read(filebuf, connection, which, bytes);

	connection->server = sock2;

	return filebuf;
}

void keepalive_check(CONNECTION * connection)
{
	if (connection->rheader->content_length == -1 && !connection->rheader->chunked) {
		/* connection can't be kept alive if we can't determine when the remote site's data ends */
		connection->keepalive_client = FALSE;
		connection->keepalive_server = FALSE;
	} else if (connection->rheader->keepalive == TRUE || connection->rheader->proxy_keepalive == TRUE)
		connection->keepalive_server = TRUE;
	else if (connection->rheader->version == HTTP_HTTP11 && (connection->rheader->keepalive != FALSE && connection->rheader->proxy_keepalive != FALSE))
		/* keep-alive is assumed for HTTP/1.1 unless any of the above conditions are met */
		connection->keepalive_server = TRUE;
	else
		connection->keepalive_server = FALSE;

	putlog(MMLOG_DEBUG, "keepalive_server = %d", connection->keepalive_server);
}

int buffer_check(CONNECTION * connection)
{
	int encode = TRUE;
	struct EXTERNAL_LIST_LIST *external_list = NULL;

	if (connection->flags & CONNECTION_PREFETCH) {
		if (connection->htmlstream == NULL || connection->rheader->content_encoding == NULL)
			return FALSE;
		else
			return TRUE;
	}

	if (connection->rheader->content_encoding != NULL && accepts_encoded_content(connection) == FALSE)
			return TRUE;

	/* check whether or not this transfer should be buffered first */
	if (rewrite_do(rewrite_list, connection, NULL, REWRITE_BODY, FALSE))
		return TRUE;

	external_list = external_find(external, connection);
	pthread_rwlock_unlock(&external->lock);

	if (external_list != NULL)
		return TRUE;

	if (keyword_check(keyword_list, connection, NULL, FALSE))
		return TRUE;

	if (url_command_find(connection->url_command, "score"))
		return TRUE;

	/* this will avoid the overhead of checking the lists again, since we've already determined none will match. */
	connection->flags &= ~CONNECTION_PROCESS;

	if (url_command_find(connection->url_command, "raw"))
		return TRUE;

	if (url_command_find(connection->url_command, "htmltree"))
		return TRUE;

	/* stream parser can't understand any encodings for now */
	if (connection->rheader->content_encoding != NULL && connection->htmlstream != NULL)
		return TRUE;

	if (atomic_read_rwlock(&global->general->lock, &global->general->compressout) == TRUE && connection->header->accept_encoding != NULL &&  connection->rheader->content_type != NULL && accepts_encoded_content(connection) && connection->rheader->content_encoding == NULL) {
		pthread_rwlock_rdlock(&global->general->lock);
		if (global->general->emp != NULL)
			encode = reg_exec(global->general->emp, connection->rheader->content_type);
		pthread_rwlock_unlock(&global->general->lock);

		if (!encode)
			return TRUE;
	}

	return FALSE;
}

int transfer_limit_check(CONNECTION * connection)
{
	// FIXME: signed/unsigned comparison
	if (connection->rheader->content_length != -1 && (connection->cachemap == NULL || (connection->flags & CONNECTION_LIMITCACHE)) && (connection->rheader->content_length > connection->transferlimit)) {
		template_send(templates, "maxtransfer", connection, 503);

		return FALSE;
	}

	return TRUE;
}

int proxy_connect(CONNECTION * connection)
{
	char *headbuf;
	HEADER *oldheader;

	oldheader = connection->header;

	connection->header = header_new();
	connection->header->type = HTTP_CONNECT;
	connection->header->host = xstrdup(oldheader->host);
	connection->header->port = oldheader->port;

	header_send(connection->header, connection, SERVER, HEADER_FORWARD);

      auth_finished:
	headbuf = header_get(connection, SERVER, atomic_read_rwlock(&global->general->lock, &global->general->timeout));

	connection->rheader = NULL;
	if (headbuf != NULL)
		connection->rheader = http_header_parse_response(headbuf);
	FREE_AND_NULL(headbuf);

	if (connection->rheader == NULL) {
		http_header_free(connection->header);
		connection->header = oldheader;

		return -1;
	}

	if (connection->rheader->code == 407 && !(connection->flags & CONNECTION_AUTHENTICATED)) {
		if (proxy_authenticate(connection)) {
			http_header_free(connection->rheader);

			connection->flags |= CONNECTION_AUTHENTICATED;

			goto auth_finished;
		}
	}

	http_header_free(connection->header);
	http_header_free(connection->rheader);

	connection->header = oldheader;
	connection->rheader = NULL;

	return TRUE;
}

void proxy_test(CONNECTION * connection)
{
	int fd, listenfd, x;
	char *headbuf, ip[16];
	struct sockaddr_in saddr;
	struct pollfd pfd;
	socklen_t sl = sizeof(struct sockaddr_in);
	SOCKET *server;

	/* bind on same interface we're connecting to the forwarding proxy on */
	getsockname(connection->server->fd, (struct sockaddr *) &saddr, &sl);
	saddr.sin_port = 0;

	listenfd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (listenfd == -1)
		return;

	x = bind(listenfd, (struct sockaddr *) &saddr, sl);
	if (x == -1) {
		close(listenfd);
		return;
	}

	x = listen(listenfd, 1);
	if (x == -1) {
		close(listenfd);
		return;
	}

	/* get the port number the kernel chose for us */
	getsockname(listenfd, (struct sockaddr *) &saddr, &sl);

	inet_ntop(AF_INET, &saddr.sin_addr, ip, sizeof(ip));
	FREE_AND_STRDUP(connection->header->host, ip);
	FREE_AND_STRDUP(connection->header->file, "/");
	connection->header->port = ntohs(saddr.sin_port);

	header_send(connection->header, connection, SERVER, HEADER_FORWARD);

	/* ok.. if all went well, the proxy should be connecting back to us */
	pfd.fd = listenfd;
	pfd.events = POLLIN;

	x = p_poll(&pfd, 1, atomic_read_rwlock(&global->general->lock, &global->general->timeout) * 1000);
	if (x <= 0) {
		close(listenfd);

		return;
	} else {
		fd = accept(listenfd, (struct sockaddr *) &saddr, &sl);
		close(listenfd);

		if (fd == -1)
			return;

		server = connection->server;
		connection->server = sock_new(fd);

		headbuf = header_get(connection, SERVER, atomic_read_rwlock(&global->general->lock, &global->general->timeout));
		if (headbuf == NULL)
			goto out;

		connection->rheader = http_header_parse_request(headbuf);
		xfree(headbuf);
	}

      out:
	sock_close(connection->server);
	connection->server = server;

	connection->keepalive_server = FALSE;
}

int accepts_encoded_content(CONNECTION *connection) {
	if (!(connection->flags & CONNECTION_ENCODED) && connection->header->accept_encoding != NULL) {
		if (s_strcasestr(connection->header->accept_encoding, "gzip") != NULL || s_strcasestr(connection->header->accept_encoding, "deflate") != NULL)
			return TRUE;
	}

	putlog(MMLOG_DEBUG, "client doesn't support encoded content");

	return FALSE;
}
